from __future__ import (
    annotations,
)

from typing import (
    Optional,
)
from uuid import (
    UUID,
)

from minos.common import (
    DatabaseMixin,
    ProgrammingException,
)

from ....exceptions import (
    SagaExecutionNotFoundException,
)
from ...saga import (
    SagaExecution,
)
from ..abc import (
    SagaExecutionRepository,
)
from .factories import (
    SagaExecutionDatabaseOperationFactory,
)


class DatabaseSagaExecutionRepository(SagaExecutionRepository, DatabaseMixin[SagaExecutionDatabaseOperationFactory]):
    """Saga Execution Storage class."""

    def __init__(self, *args, database_key: Optional[tuple[str]] = None, **kwargs):
        if database_key is None:
            database_key = ("saga",)
        super().__init__(*args, database_key=database_key, **kwargs)

    async def _setup(self) -> None:
        await super()._setup()
        operation = self.database_operation_factory.build_create()
        await self.execute_on_database(operation)

    async def _store(self, execution: SagaExecution) -> None:
        operation = self.database_operation_factory.build_store(**execution.raw)
        await self.execute_on_database(operation)

    async def _delete(self, uuid: UUID) -> None:
        operation = self.database_operation_factory.build_delete(uuid)
        await self.execute_on_database(operation)

    async def _load(self, uuid: UUID) -> SagaExecution:
        operation = self.database_operation_factory.build_load(uuid)

        try:
            value = await self.execute_on_database_and_fetch_one(operation)
        except ProgrammingException:
            raise SagaExecutionNotFoundException(f"The execution identified by {uuid} was not found.")
        execution = SagaExecution.from_raw(value)
        return execution



from collections.abc import (
    MutableMapping,
)
from typing import (
    Any,
)

from minos.common import (
    BucketModel,
    Field,
)


class SagaContext(BucketModel, MutableMapping):
    """Saga Context class

    The purpose of this class is to keep an execution state.
    """

    def __init__(self, **kwargs):
        if "fields" not in kwargs:
            fields = {name: Field(name, Any, value) for name, value in kwargs.items()}
        else:
            fields = {name: Field(field.name, Any, field.value) for name, field in kwargs["fields"].items()}

        super().__init__(fields=fields)

    def __setitem__(self, key: str, value: Any) -> None:
        try:
            super().__setitem__(key, value)
        except KeyError:
            self._fields[key] = Field(key, Any, value)

    def __delitem__(self, item: str) -> None:
        try:
            del self._fields[item]
        except KeyError:
            raise KeyError(f"{type(self).__name__!r} does not contain the {item!r} field")

    def __setattr__(self, key: str, value: Any) -> None:
        try:
            super().__setattr__(key, value)
        except AttributeError:
            self[key] = value

    def __delattr__(self, item: str) -> None:
        if item.startswith("_"):
            super().__delattr__(item)
            return

        try:
            del self[item]
        except KeyError as exc:
            raise AttributeError(str(exc))



"""Runners module."""

from __future__ import (
    annotations,
)

import logging
import warnings
from asyncio import (
    TimeoutError,
    wait_for,
)
from functools import (
    reduce,
)
from operator import (
    or_,
)
from typing import (
    Optional,
    Union,
)
from uuid import (
    UUID,
)

from minos.common import (
    Inject,
    NotProvidedException,
    PoolFactory,
    SetupMixin,
)
from minos.networks import (
    REQUEST_HEADERS_CONTEXT_VAR,
    REQUEST_USER_CONTEXT_VAR,
    BrokerClient,
    BrokerClientPool,
    BrokerMessage,
)

from ..context import (
    SagaContext,
)
from ..definitions import (
    Saga,
    SagaDecoratorWrapper,
)
from ..exceptions import (
    SagaFailedExecutionException,
    SagaPausedExecutionStepException,
)
from ..messages import (
    SagaResponse,
)
from .repositories import (
    SagaExecutionRepository,
)
from .saga import (
    SagaExecution,
    SagaStatus,
)

logger = logging.getLogger(__name__)


class SagaRunner(SetupMixin):
    """Saga Runner class."""

    @Inject()
    def __init__(
        self,
        storage: SagaExecutionRepository,
        broker_pool: Optional[BrokerClientPool] = None,
        pool_factory: Optional[PoolFactory] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if storage is None:
            raise NotProvidedException(f"A {SagaExecutionRepository!r} instance is required.")

        if broker_pool is None and pool_factory is not None:
            broker_pool = pool_factory.get_pool("broker")

        if broker_pool is None:
            raise NotProvidedException(f"A {BrokerClientPool!r} instance is required.")

        self.storage = storage
        self.broker_pool = broker_pool

    async def run(
        self,
        definition: Optional[Union[Saga, SagaDecoratorWrapper]] = None,
        context: Optional[SagaContext] = None,
        *,
        response: Optional[SagaResponse] = None,
        user: Optional[UUID] = None,
        autocommit: bool = True,
        pause_on_disk: bool = False,
        raise_on_error: bool = True,
        return_execution: bool = True,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Union[UUID, SagaExecution]:
        """Perform a run of a ``Saga``.

        The run can be a new one (if a name is provided) or continue execution a previous one (if a reply is provided).

        :param definition: Saga definition to be executed.
        :param context: Initial context to be used during the execution. (Only used for new executions)
        :param response: The reply that relaunches a saga execution.
        :param user: The user identifier to be injected on remote steps.
        :param autocommit: If ``True`` the transactions are committed/rejected automatically. Otherwise, the ``commit``
            or ``reject`` must be called manually.
        :param pause_on_disk: If ``True`` the pauses until remote steps' responses are paused on disk (background,
            non-blocking the execution). Otherwise, the pauses are waited on memory (online, blocking the execution)
        :param raise_on_error: If ``True`` exceptions are raised on error. Otherwise, the execution is returned normally
            but with ``Errored`` status.
        :param return_execution: If ``True`` the ``SagaExecution`` instance is returned. Otherwise, only the
            identifier (``UUID``) is returned.
        :param timeout: Maximum execution time in seconds.
        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """
        if isinstance(definition, SagaDecoratorWrapper):
            definition = definition.meta.definition

        if response is None:
            execution = await self._create(definition, context, user)
        else:
            execution = await self._load(response)

        execution = await self._run(
            execution,
            timeout=timeout,
            response=response,
            autocommit=autocommit,
            pause_on_disk=pause_on_disk,
            raise_on_error=raise_on_error,
            **kwargs,
        )
        if return_execution:
            return execution

        return execution.uuid

    @staticmethod
    async def _create(definition: Saga, context: Optional[SagaContext], user: Optional[UUID]) -> SagaExecution:
        if REQUEST_USER_CONTEXT_VAR.get() is not None:
            if user is not None:
                warnings.warn("The `user` Argument will be ignored in favor of the `user` ContextVar", RuntimeWarning)
            user = REQUEST_USER_CONTEXT_VAR.get()

        return SagaExecution.from_definition(definition, context=context, user=user)

    async def _load(self, response: SagaResponse) -> Union[UUID, SagaExecution]:
        return await self.storage.load(response.uuid)

    async def _run(self, execution: SagaExecution, raise_on_error: bool, **kwargs) -> SagaExecution:
        try:
            await self._run_with_timeout(execution, **kwargs)
        except SagaFailedExecutionException as exc:
            if raise_on_error:
                raise exc
            logger.exception(f"The execution identified by {execution.uuid!s} failed: {exc.exception!r}")
        finally:
            await self.storage.store(execution)
            self._update_request_headers(execution)

        return execution

    async def _run_with_timeout(self, execution: SagaExecution, timeout: Optional[float], **kwargs) -> None:
        future = self._run_with_pause(execution, **kwargs)
        try:
            return await wait_for(future, timeout=timeout)
        except TimeoutError as exc:
            raise SagaFailedExecutionException(exc)

    async def _run_with_pause(self, execution: SagaExecution, pause_on_disk: bool, **kwargs) -> None:
        if pause_on_disk:
            await self._run_with_pause_on_disk(execution, **kwargs)
        else:
            await self._run_with_pause_on_memory(execution, **kwargs)

    @staticmethod
    def _update_request_headers(execution: SagaExecution) -> None:
        if (headers := REQUEST_HEADERS_CONTEXT_VAR.get()) is None:
            return

        related_services = reduce(or_, (s.related_services for s in execution.executed_steps), set())
        if execution.paused_step is not None:
            related_services.update(execution.paused_step.related_services)

        if raw_related_services := headers.get("related_services"):
            related_services.update(raw_related_services.split(","))

        headers["related_services"] = ",".join(related_services)

    @staticmethod
    async def _run_with_pause_on_disk(
        execution: SagaExecution, response: Optional[SagaResponse] = None, autocommit: bool = True, **kwargs
    ) -> None:
        try:
            await execution.execute(autocommit=False, response=response, **kwargs)
            if autocommit:
                await execution.commit(**kwargs)
        except SagaPausedExecutionStepException:
            return
        except SagaFailedExecutionException as exc:
            if autocommit:
                await execution.reject(**kwargs)
            raise exc

    async def _run_with_pause_on_memory(
        self, execution: SagaExecution, response: Optional[SagaResponse] = None, autocommit: bool = True, **kwargs
    ) -> None:

        try:
            # noinspection PyUnresolvedReferences
            async with self.broker_pool.acquire() as broker:
                while execution.status in (SagaStatus.Created, SagaStatus.Paused):
                    try:
                        await execution.execute(response=response, autocommit=False, **kwargs)
                    except SagaPausedExecutionStepException:
                        await self.storage.store(execution)
                        response = await self._get_response(broker, execution, **kwargs)
            if autocommit:
                await execution.commit(**kwargs)
        except SagaFailedExecutionException as exc:
            if autocommit:
                await execution.reject(**kwargs)
            raise exc

    @staticmethod
    async def _get_response(broker: BrokerClient, execution: SagaExecution, **kwargs) -> SagaResponse:
        message: Optional[BrokerMessage] = None
        while message is None or "saga" not in message.headers or UUID(message.headers["saga"]) != execution.uuid:
            try:
                message = await broker.receive(**kwargs)
            except Exception as exc:
                execution.status = SagaStatus.Errored
                raise SagaFailedExecutionException(exc)
        return SagaResponse.from_message(message)



from collections.abc import (
    Awaitable,
    Callable,
)
from typing import (
    Optional,
)
from uuid import (
    UUID,
)

from minos.networks import (
    REQUEST_HEADERS_CONTEXT_VAR,
    BrokerRequest,
    Request,
    Response,
)
from minos.transactions import (
    TransactionEntry,
)

from .utils import (
    get_service_name,
)


async def transactional_command(
    request: Request, inner: Callable[[Request], Awaitable[Optional[Response]]]
) -> Optional[Response]:
    """Execute the command transactionally if it comes from a saga.

    :param request: The request containing the data.
    :param inner: The inner handling function to be executed.
    :return: The response generated by the inner handling function.
    """
    try:
        if isinstance(request, BrokerRequest):
            message = request.raw
            if raw_transaction_uuids := message.headers.get("transactions"):
                transaction_uuids = list(map(UUID, raw_transaction_uuids.split(",")))
                return await _transaction(request, inner, transaction_uuids)
        return await inner(request)
    finally:
        if (headers := REQUEST_HEADERS_CONTEXT_VAR.get()) is not None:
            if (raw_transaction_uuids := headers.get("transactions")) is not None:
                raw_transaction_uuids_parts = raw_transaction_uuids.rsplit(",", 1)
                if len(raw_transaction_uuids_parts) == 1:
                    del headers["transactions"]
                else:
                    headers["transactions"] = raw_transaction_uuids_parts[0]

            related_services = set()
            if raw_related_services := headers.get("related_services"):
                related_services.update(raw_related_services.split(","))
            related_services.add(get_service_name())
            headers["related_services"] = ",".join(related_services)


async def _transaction(
    request: Request, inner: Callable[[Request], Awaitable[Optional[Response]]], transaction_uuids: list[UUID]
) -> Optional[Response]:
    if len(transaction_uuids):
        async with TransactionEntry(uuid=transaction_uuids[0], autocommit=False):
            return await _transaction(request, inner, transaction_uuids[1:])

    return await inner(request)



from __future__ import (
    annotations,
)

from enum import (
    IntEnum,
)
from typing import (
    Any,
    Optional,
    Union,
)
from uuid import (
    UUID,
)

from minos.networks import (
    BrokerMessage,
)


class SagaRequest:
    """Saga Request class."""

    __slots__ = (
        "_target",
        "_content",
    )

    def __init__(self, target: str, content: Any = None):
        self._target = target
        self._content = content

    @property
    def target(self) -> str:
        """Get the target of the request.

        :return: A ``str`` instance.
        """
        return self._target

    # noinspection PyUnusedLocal
    async def content(self, **kwargs) -> Any:
        """Get the content of the request.

        :param kwargs: Additional named parameters.
        :return: The content of the request.
        """
        return self._content

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and self._target == other._target and self._content == other._content

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._target!r}, {self._content!r})"

    def __hash__(self):
        return hash((self._target, self._content))


class SagaResponse:
    """Saga Response class."""

    __slots__ = (
        "_content",
        "_status",
        "_related_services",
        "_uuid",
    )

    def __init__(
        self,
        content: Any = None,
        related_services: Optional[set[str]] = None,
        status: Optional[Union[int, SagaResponseStatus]] = None,
        uuid: Optional[UUID] = None,
        *args,
        **kwargs,
    ):
        if status is None:
            status = SagaResponseStatus.SUCCESS
        if not isinstance(status, SagaResponseStatus):
            status = SagaResponseStatus.from_raw(status)
        if related_services is None:
            related_services = set()

        self._content = content
        self._status = status
        self._related_services = related_services
        self._uuid = uuid

    @classmethod
    def from_message(cls, message: BrokerMessage) -> SagaResponse:
        """Build a new ``SagaResponse`` from a ``BrokerMessage``.

        :param message: The ``BrokerMessage`` instance.
        :return: A ``SagaResponse``.
        """

        uuid = UUID(message.headers["saga"])

        if raw_related_services := message.headers.get("related_services"):
            related_services = set(raw_related_services.split(","))
        else:
            related_services = set()

        return SagaResponse(message.content, related_services, message.status, uuid)

    # noinspection PyUnusedLocal
    async def content(self, **kwargs) -> Any:
        """Get the response content.

        :param kwargs: Additional named parameters.
        :return: The content of the response.
        """
        return self._content

    @property
    def ok(self) -> bool:
        """Check if the response is okay.

        :return: ``True`` if the response is okay
        """
        return self._status == SagaResponseStatus.SUCCESS

    @property
    def status(self) -> SagaResponseStatus:
        """Get the status code of the response.

        :return: A ``ResponseStatus`` instance.
        """
        return self._status

    @property
    def related_services(self) -> set[str]:
        """Get the microservice name that generated the response.

        :return: An string value containing the microservice name.
        """
        return self._related_services

    @property
    def uuid(self) -> UUID:
        """Get the identifier of the saga execution that must receive the response.

        :return: An ``UUID`` value.
        """
        return self._uuid

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, type(self))
            and self._content == other._content
            and self._status == other._status
            and self._related_services == other._related_services
            and self._uuid == other._uuid
        )

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"{self._content!r}, {self._status!r}, {self._related_services!r}, {self._uuid!r}"
            f")"
        )

    def __hash__(self):
        return hash((self._content, self._status, tuple(sorted(self._related_services)), self._uuid))


class SagaResponseStatus(IntEnum):
    """Saga Response Status class."""

    SUCCESS = 200
    ERROR = 400
    SYSTEM_ERROR = 500

    @classmethod
    def from_raw(cls, raw: int) -> SagaResponseStatus:
        """Build a new instance from raw.

        :param raw: The raw representation of the instance.
        :return: A ``SagaResponseStatus`` instance.
        """
        return next(obj for obj in cls if obj.value == raw)



from ...context import (
    SagaContext,
)
from ...definitions import (
    LocalSagaStep,
)
from ...exceptions import (
    SagaFailedExecutionStepException,
    SagaRollbackExecutionStepException,
)
from ...utils import (
    get_service_name,
)
from ..executors import (
    LocalExecutor,
)
from ..status import (
    SagaStepStatus,
)
from .abc import (
    SagaStepExecution,
)


class LocalSagaStepExecution(SagaStepExecution):
    """Local Saga Step Execution class."""

    definition: LocalSagaStep

    async def execute(self, context: SagaContext, *args, **kwargs) -> SagaContext:
        """Execute the local saga step.

        :param context: The execution context.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: The new saga context.
        """

        if self.status != SagaStepStatus.Created:
            return context

        context = await self._execute_on_execute(context, *args, **kwargs)

        self.status = SagaStepStatus.Finished
        return context

    async def _execute_on_execute(self, context: SagaContext, *args, **kwargs) -> SagaContext:

        self.status = SagaStepStatus.RunningOnExecute

        executor = LocalExecutor(*args, **kwargs)

        self.related_services.add(get_service_name())
        try:
            context = await executor.exec(self.definition.on_execute_operation, context)
        except SagaFailedExecutionStepException as exc:
            # await self.rollback(*args, **kwargs)  # Rollback must not be performed at this point.
            self.status = SagaStepStatus.ErroredOnExecute
            raise exc

        self.status = SagaStepStatus.FinishedOnExecute

        return context

    async def rollback(self, context: SagaContext, *args, **kwargs) -> SagaContext:
        """Rollback the local saga context.

        :param context: The execution context.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: The new saga context.
        """

        if self.status == SagaStepStatus.Created:
            raise SagaRollbackExecutionStepException("There is nothing to rollback.")

        if self.already_rollback:
            raise SagaRollbackExecutionStepException("The step was already rollbacked.")

        executor = LocalExecutor(*args, **kwargs)
        self.related_services.add(get_service_name())
        context = await executor.exec(self.definition.on_failure_operation, context)

        self.already_rollback = True
        return context



from abc import (
    ABC,
    abstractmethod,
)
from uuid import (
    UUID,
)

from minos.common import (
    DatabaseOperation,
    DeclarativeModel,
)
from minos.common.testing import (
    MinosTestCase,
    MockedDatabaseClient,
    MockedDatabaseOperation,
)

from .context import (
    SagaContext,
)
from .definitions import (
    Saga,
)
from .exceptions import (
    SagaExecutionNotFoundException,
)
from .executions import (
    SagaExecution,
    SagaExecutionDatabaseOperationFactory,
    SagaExecutionRepository,
)


class MockedSagaExecutionDatabaseOperationFactory(SagaExecutionDatabaseOperationFactory):
    """For testing purposes"""

    def build_create(self) -> DatabaseOperation:
        """For testing purposes"""
        return MockedDatabaseOperation("create_table")

    def build_store(self, uuid: UUID, **kwargs) -> DatabaseOperation:
        """For testing purposes"""
        return MockedDatabaseOperation("create_table")

    def build_load(self, uuid: UUID) -> DatabaseOperation:
        """For testing purposes"""
        return MockedDatabaseOperation("create_table")

    def build_delete(self, uuid: UUID) -> DatabaseOperation:
        """For testing purposes"""
        return MockedDatabaseOperation("create_table")


MockedDatabaseClient.set_factory(
    SagaExecutionDatabaseOperationFactory,
    MockedSagaExecutionDatabaseOperationFactory,
)


class Foo(DeclarativeModel):
    """For testing purposes."""

    foo: str


def _fn1(context: SagaContext) -> SagaContext:  # pragma: no cover
    context["payment"] = "payment"
    return context


def _fn2(context: SagaContext) -> SagaContext:  # pragma: no cover
    context["payment"] = None
    return context


_SAGA = Saga().local_step(_fn1).on_failure(_fn2).commit()


class SagaExecutionRepositoryTestCase(MinosTestCase, ABC):
    __test__ = False

    def setUp(self) -> None:
        super().setUp()
        self.saga_execution_repository = self.build_saga_execution_repository()
        self.execution = SagaExecution.from_definition(_SAGA)
        self.another = SagaExecution.from_definition(_SAGA)

    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()
        await self.saga_execution_repository.setup()
        await self.execution.execute(autocommit=False)

    async def asyncTearDown(self):
        await self.saga_execution_repository.destroy()
        await super().asyncTearDown()

    @abstractmethod
    def build_saga_execution_repository(self) -> SagaExecutionRepository:
        """For testing purposes."""

    async def test_store(self):
        await self.saga_execution_repository.store(self.execution)

        self.assertEqual(self.execution, await self.saga_execution_repository.load(self.execution.uuid))

    async def test_store_overwrite(self):
        await self.saga_execution_repository.store(self.execution)
        self.assertEqual(self.execution, await self.saga_execution_repository.load(self.execution.uuid))

        self.another.uuid = self.execution.uuid
        await self.saga_execution_repository.store(self.another)

        self.assertNotEqual(self.execution, await self.saga_execution_repository.load(self.execution.uuid))
        self.assertEqual(self.another, await self.saga_execution_repository.load(self.execution.uuid))

    async def test_load_from_str(self):
        await self.saga_execution_repository.store(self.execution)
        self.assertEqual(self.execution, await self.saga_execution_repository.load(str(self.execution.uuid)))

    async def test_load_raises(self):
        with self.assertRaises(SagaExecutionNotFoundException):
            await self.saga_execution_repository.load(self.execution.uuid)

    async def test_delete(self):
        await self.saga_execution_repository.store(self.execution)
        await self.saga_execution_repository.delete(self.execution)
        with self.assertRaises(SagaExecutionNotFoundException):
            await self.saga_execution_repository.load(self.execution.uuid)

    async def test_delete_from_str(self):
        await self.saga_execution_repository.store(self.execution)
        await self.saga_execution_repository.delete(str(self.execution.uuid))
        with self.assertRaises(SagaExecutionNotFoundException):
            await self.saga_execution_repository.load(self.execution.uuid)



import logging
from datetime import (
    timedelta,
)

from minos.common import (
    UUID_REGEX,
    Config,
    Inject,
    current_datetime,
)
from minos.networks import (
    BrokerRequest,
    EnrouteDecorator,
    Request,
    Response,
    ResponseException,
    enroute,
)
from minos.saga import (
    SagaManager,
    SagaResponse,
    SagaStatus,
)
from minos.transactions import (
    TransactionNotFoundException,
    TransactionRepository,
    TransactionRepositoryConflictException,
    TransactionStatus,
)

logger = logging.getLogger(__name__)


class SagaService:
    """Saga Service class"""

    # noinspection PyUnusedLocal
    @Inject()
    def __init__(self, *args, saga_manager: SagaManager, transaction_repository: TransactionRepository, **kwargs):
        self.saga_manager = saga_manager
        self.transaction_repository = transaction_repository

    @classmethod
    def __get_enroute__(cls, config: Config) -> dict[str, set[EnrouteDecorator]]:
        name = config.get_name()
        return {
            cls._handle_reply.__name__: {
                enroute.broker.command(f"{config.get_name()}Reply"),
            },
            cls._handle_get.__name__: {
                enroute.broker.command(f"_Get{name.title()}SagaExecution"),
                enroute.rest.command(f"/{name}s/saga/executions/{{uuid:{UUID_REGEX.pattern}}}", "GET"),
            },
            cls._handle_reserve.__name__: {
                enroute.broker.command(f"_Reserve{name.title()}Transaction"),
            },
            cls._handle_reject.__name__: {
                enroute.broker.command(f"_Reject{name.title()}Transaction"),
            },
            cls._handle_commit.__name__: {
                enroute.broker.command(f"_Commit{name.title()}Transaction"),
            },
            cls._handle_reject_blocked.__name__: {
                enroute.periodic.event("* * * * *"),
            },
        }

    async def _handle_reply(self, request: BrokerRequest) -> None:
        response = SagaResponse.from_message(request.raw)
        await self.saga_manager.run(response=response, pause_on_disk=True, raise_on_error=False, return_execution=False)

    async def _handle_get(self, request: Request) -> Response:
        if request.has_params:
            uuid = (await request.params())["uuid"]
        else:
            uuid = (await request.content())["uuid"]

        execution = await self.saga_manager.get(uuid)

        content = {"uuid": execution.uuid, "status": execution.status}
        if execution.status == SagaStatus.Finished:
            content["context"] = execution.context

        return Response(content)

    async def _handle_reserve(self, request: Request) -> None:
        uuid = await request.content()

        try:
            transaction = await self.transaction_repository.get(uuid)
        except TransactionNotFoundException:
            raise ResponseException(f"The transaction identified by {uuid!r} does not exist.")

        try:
            await transaction.reserve()
        except TransactionRepositoryConflictException:
            raise ResponseException("The transaction could not be reserved.")

    async def _handle_reject(self, request: Request) -> None:
        uuid = await request.content()

        try:
            transaction = await self.transaction_repository.get(uuid)
        except TransactionNotFoundException:
            raise ResponseException(f"The transaction identified by {uuid!r} does not exist.")

        if transaction.status == TransactionStatus.REJECTED:
            return

        try:
            await transaction.reject()
        except TransactionRepositoryConflictException as exc:
            raise ResponseException(f"{exc!s}")

    async def _handle_commit(self, request: Request) -> None:
        uuid = await request.content()

        try:
            transaction = await self.transaction_repository.get(uuid)
        except TransactionNotFoundException:
            raise ResponseException(f"The transaction identified by {uuid!r} does not exist.")

        try:
            await transaction.commit()
        except TransactionRepositoryConflictException as exc:
            raise ResponseException(f"{exc!s}")

    # noinspection PyUnusedLocal
    async def _handle_reject_blocked(self, request: Request) -> None:
        status_in = (TransactionStatus.RESERVED,)
        updated_at_lt = current_datetime() - timedelta(minutes=1)
        async for transaction in self.transaction_repository.select(status_in=status_in, updated_at_lt=updated_at_lt):
            logger.info(f"Rejecting {transaction.uuid!r} transaction as it has been reserved for a long time...")
            try:
                await transaction.reject()
            except TransactionRepositoryConflictException as exc:
                logger.warning(f"Raised an exception while trying to reject a blocked transaction: {exc!r}")



"""Saga definitions module."""

from __future__ import (
    annotations,
)

import warnings
from collections.abc import (
    Iterable,
)
from inspect import (
    getmembers,
)
from operator import (
    attrgetter,
)
from typing import (
    Any,
    Optional,
    Protocol,
    TypeVar,
    Union,
    runtime_checkable,
)

from cached_property import (
    cached_property,
)

from ..exceptions import (
    AlreadyCommittedException,
    AlreadyOnSagaException,
    EmptySagaException,
    OrderPrecedenceException,
    SagaNotCommittedException,
)
from .operations import (
    SagaOperation,
)
from .steps import (
    ConditionalSagaStep,
    LocalSagaStep,
    RemoteSagaStep,
    SagaStep,
    SagaStepDecoratorWrapper,
)
from .types import (
    LocalCallback,
    RequestCallBack,
)


@runtime_checkable
class SagaDecoratorWrapper(Protocol):
    """Saga Decorator Wrapper class."""

    meta: SagaDecoratorMeta


class SagaDecoratorMeta:
    """Saga Decorator Meta class."""

    _inner: type
    _definition: Saga

    def __init__(self, func: type, saga: Saga):
        self._inner = func
        self._definition = saga

    @cached_property
    def definition(self) -> Saga:
        """Get the saga definition.

        :return: A ``Saga`` instance.
        """
        steps = getmembers(self._inner, predicate=lambda x: isinstance(x, SagaStepDecoratorWrapper))
        steps = list(map(lambda member: member[1].meta.definition, steps))
        for step in steps:
            if step.order is None:
                raise OrderPrecedenceException(f"The {step!r} step does not have 'order' value.")
        steps.sort(key=attrgetter("order"))

        for step in steps:
            self._definition.add_step(step)
        self._definition.commit()

        return self._definition


TP = TypeVar("TP", bound=type)


class Saga:
    """Saga class.

    The purpose of this class is to define a sequence of operations among microservices.
    """

    # noinspection PyUnusedLocal
    def __init__(
        self, *args, steps: list[SagaStep] = None, committed: Optional[bool] = None, commit: None = None, **kwargs
    ):
        self.steps = list()
        self.committed = False

        if steps is None:
            steps = list()
        for step in steps:
            self.add_step(step)

        if committed is None:
            committed = len(steps)
        self.committed = committed

        if commit is not None:
            warnings.warn(f"Commit callback is being deprecated. Use {self.local_step!r} instead", DeprecationWarning)
            self.local_step(commit)
            self.committed = True

    def __call__(self, type_: TP) -> Union[TP, SagaDecoratorWrapper]:
        """Decorate the given type.

        :param type_: The type to be decorated.
        :return: The decorated type.
        """
        type_.meta = SagaDecoratorMeta(type_, self)
        return type_

    @classmethod
    def from_raw(cls, raw: Union[dict[str, Any], Saga], **kwargs) -> Saga:
        """Build a new ``Saga`` instance from raw.

        :param raw: The raw representation of the saga.
        :param kwargs: Additional named arguments.
        :return: A new ``Saga`` instance.
        """
        if isinstance(raw, cls):
            return raw

        current = raw | kwargs

        steps = [SagaStep.from_raw(step) for step in current.pop("steps")]

        instance = cls(steps=steps, **current)

        return instance

    def conditional_step(self, step: Optional[ConditionalSagaStep] = None) -> ConditionalSagaStep:
        """Add a new conditional step.

        :param step: The step to be added. If `None` is provided then a new one will be created.
        :return: A ``SagaStep`` instance.
        """
        return self.add_step(step, ConditionalSagaStep)

    def local_step(
        self, step: Optional[Union[LocalCallback, SagaOperation[LocalCallback], LocalSagaStep]] = None, **kwargs
    ) -> LocalSagaStep:
        """Add a new local step.

        :param step: The step to be added. If `None` is provided then a new one will be created.
        :param kwargs: Additional named parameters.
        :return: A ``SagaStep`` instance.
        """
        if step is not None and not isinstance(step, SagaStep) and not isinstance(step, SagaOperation):
            step = SagaOperation(step, **kwargs)
        return self.add_step(step, LocalSagaStep)

    def step(
        self, step: Optional[Union[RequestCallBack, SagaOperation[RequestCallBack], RemoteSagaStep]] = None, **kwargs
    ) -> RemoteSagaStep:
        """Add a new remote step.

        :param step: The step to be added. If `None` is provided then a new one will be created.
        :param kwargs: Additional named parameters.
        :return: A ``SagaStep`` instance.
        """
        warnings.warn("step() method is deprecated by remote_step() and will be removed soon.", DeprecationWarning)
        return self.remote_step(step, **kwargs)

    def remote_step(
        self, step: Optional[Union[RequestCallBack, SagaOperation[RequestCallBack], RemoteSagaStep]] = None, **kwargs
    ) -> RemoteSagaStep:
        """Add a new remote step.

        :param step: The step to be added. If `None` is provided then a new one will be created.
        :param kwargs: Additional named parameters.
        :return: A ``SagaStep`` instance.
        """
        if step is not None and not isinstance(step, SagaStep) and not isinstance(step, SagaOperation):
            step = SagaOperation(step, **kwargs)
        return self.add_step(step, RemoteSagaStep)

    def add_step(self, step: Optional[Union[SagaOperation, T]], step_cls: type[T] = SagaStep) -> T:
        """Add a new step.

        :param step: The step to be added.
        :param step_cls: The step class (for validation purposes).
        :return: The added step.
        """
        if self.committed:
            raise AlreadyCommittedException("It is not possible to add more steps to an already committed saga.")

        if step is None:
            step = step_cls(saga=self)
        elif isinstance(step, SagaStep):
            if not isinstance(step, step_cls):
                raise TypeError(f"The given step does not match the requested type: {step_cls} vs {type(step)}")
            if step.saga is not None:
                raise AlreadyOnSagaException()
            step.saga = self

        else:
            step = step_cls(step, saga=self)

        if step.order is None:
            if self.steps:
                step.order = self.steps[-1].order + 1
            else:
                step.order = 1

        if self.steps and step.order <= self.steps[-1].order:
            raise OrderPrecedenceException(
                f"Unsatisfied precedence constraints. Previous: {self.steps[-1].order} Current: {step.order} "
            )

        self.steps.append(step)
        return step

    @property
    def raw(self) -> dict[str, Any]:
        """Generate a raw representation of the instance.

        :return: A ``dict`` instance.
        """
        ans = {
            "steps": [step.raw for step in self.steps],
            "committed": self.committed,
        }
        return ans

    def __eq__(self, other: SagaStep) -> bool:
        return type(self) == type(other) and tuple(self) == tuple(other)

    def __iter__(self) -> Iterable:
        yield from (
            self.steps,
            self.committed,
        )

    def commit(self, callback: None = None, **kwargs) -> Saga:
        """Commit the instance to be ready for execution.

        :param callback: Deprecated argument.
        :param kwargs: Additional named arguments.
        :return: A ``Saga`` instance.
        """
        if self.committed:
            raise AlreadyCommittedException("It is not possible to commit a saga multiple times.")

        if callback is not None:
            warnings.warn(f"Commit callback is being deprecated. Use {self.local_step!r} instead", DeprecationWarning)
            self.local_step(callback, **kwargs)

        self.committed = True
        return self

    def validate(self) -> None:
        """Check if the saga is valid.

        :return: This method does not return anything, but raises an exception if the saga is not valid.
        """

        if not len(self.steps):
            raise EmptySagaException()

        for step in self.steps:
            step.validate()

        if not self.committed:
            raise SagaNotCommittedException()


T = TypeVar("T")



from minos.common import (
    MinosException,
)


class SagaException(MinosException):
    """Base saga exception."""


class EmptySagaException(SagaException):
    """Exception to be raised when saga is empty."""

    def __init__(self, message: str = None):
        if message is None:
            message = "A 'Saga' must have at least one step."
        super().__init__(message)


class SagaNotCommittedException(SagaException):
    """Exception to be raised when trying to exec a  not committed saga."""

    def __init__(self, message: str = None):
        if message is None:
            message = "A 'Saga' must be committed."

        super().__init__(message)


class SagaStepException(SagaException):
    """Base exception for saga steps."""


class SagaNotDefinedException(SagaStepException):
    """Exception to be raised when the saga is not defined."""

    def __init__(self, message: str = None):
        if message is None:
            message = "A 'SagaStep' must have a 'Saga' instance to call call this method."
        super().__init__(message)


class EmptySagaStepException(SagaStepException):
    """Exception to be raised when the step is empty."""

    def __init__(self, message: str = None):
        if message is None:
            message = "A 'SagaStep' must have at least one defined action."
        super().__init__(message)


class MultipleOnExecuteException(SagaStepException):
    """Exception to be raised when multiple on execute methods are defined."""

    def __init__(self, message: str = None):
        if message is None:
            message = "A 'SagaStep' can only define one 'on_execute' method."
        super().__init__(message)


class MultipleOnFailureException(SagaStepException):
    """Exception to be raised when multiple on failure methods are defined."""

    def __init__(self, message: str = None):
        if message is None:
            message = "A 'SagaStep' can only define one 'on_failure' method."
        super().__init__(message)


class MultipleOnSuccessException(SagaStepException):
    """Exception to be raised when multiple on success methods are defined."""

    def __init__(self, message: str = None):
        if message is None:
            message = "A 'SagaStep' can only define one 'on_success' method."
        super().__init__(message)


class MultipleOnErrorException(SagaStepException):
    """Exception to be raised when multiple on error methods are defined."""

    def __init__(self, message: str = None):
        if message is None:
            message = "A 'SagaStep' can only define one 'on_error' method."
        super().__init__(message)


class MultipleElseThenException(SagaStepException):
    """Exception to be raised when multiple else then alternatives are defined."""

    def __init__(self, message: str = None):
        if message is None:
            message = "A 'ConditionalSagaStep' can only define one 'else_then' method."
        super().__init__(message)


class AlreadyOnSagaException(SagaStepException):
    """Exception to be raised when a saga step is already in another saga."""

    def __init__(self, message: str = None):
        if message is None:
            message = "A 'SagaStep' can only belong to one 'Saga' simultaneously."
        super().__init__(message)


class UndefinedOnExecuteException(SagaStepException):
    """Exception to be raised when the on execute method is not defined."""

    def __init__(self, message: str = None):
        if message is None:
            message = "A 'SagaStep' must define at least the 'on_execute' logic."
        super().__init__(message)


class SagaExecutionException(SagaException):
    """Base exception for saga execution."""


class SagaExecutionNotFoundException(SagaExecutionException):
    """Exception to be raised when a saga execution is not found."""


class SagaRollbackExecutionException(SagaExecutionException):
    """Exception to be raised when a saga exception cannot be rollbacked"""


class SagaFailedExecutionException(SagaExecutionException):
    """Exception to be raised when a saga execution failed while running."""

    def __init__(self, exception: Exception, message: str = None):
        self.exception = exception
        if message is None:
            message = f"There was a failure while 'SagaStepExecution' was executing: {exception!r}"
        super().__init__(message)


class SagaExecutionAlreadyExecutedException(SagaExecutionException):
    """Exception to be raised when a saga execution cannot be executed."""


class SagaStepExecutionException(SagaException):
    """Base exception for saga execution step."""


class SagaFailedExecutionStepException(SagaStepExecutionException, SagaFailedExecutionException):
    """Exception to be raised when a saga execution step failed while running."""


class SagaPausedExecutionStepException(SagaStepExecutionException):
    """Exception to be raised when a saga execution step is paused."""

    def __init__(self, message: str = None):
        if message is None:
            message = "There was a pause while 'SagaStepExecution' was executing."
        super().__init__(message)


class SagaRollbackExecutionStepException(SagaStepExecutionException):
    """Exception to be raised when a saga execution step failed while performing a rollback."""


class AlreadyCommittedException(SagaException):
    """Exception to be raised when trying to modify an already committed saga."""


class OrderPrecedenceException(SagaException):
    """Exception to be raised when ordering precedence of steps is not satisfied."""


class ExecutorException(SagaException):
    """Exception to be raised when a saga executor raises some exception."""

    def __init__(self, exception: Exception, message: str = None):
        self.exception = exception
        if message is None:
            message = f"There was a failure while 'SagaStepExecution' was executing: {exception!r}"
        super().__init__(message)


class SagaFailedCommitCallbackException(SagaFailedExecutionException):
    """Exception to be raised when a saga commit callback raises some exception"""


class SagaResponseException(SagaException):
    """Exception to be used when ``CommandStatus`` is not ``SUCCESS``"""



from typing import (
    Optional,
)

from ...context import (
    SagaContext,
)
from ...definitions import (
    LocalCallback,
    SagaOperation,
)
from ...exceptions import (
    ExecutorException,
    SagaFailedExecutionStepException,
)
from .abc import (
    Executor,
)


class LocalExecutor(Executor):
    """Local Executor class."""

    # noinspection PyUnusedLocal,PyMethodOverriding
    async def exec(
        self, operation: Optional[SagaOperation[LocalCallback]], context: SagaContext, *args, **kwargs
    ) -> SagaContext:
        """Execute the commit operation.

        :param operation: Operation to be executed.
        :param context: Actual execution context.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: An updated context instance.
        """

        if operation is None:
            return context

        try:
            context = SagaContext(**context)  # Needed to avoid mutability issues.
            new_context = await super().exec(operation, context)
        except ExecutorException as exc:
            raise SagaFailedExecutionStepException(exc.exception)

        if new_context is not None:
            context = new_context

        return context



from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    Iterable,
    Optional,
    Union,
)

from minos.common import (
    classname,
    import_module,
)

from ...context import (
    SagaContext,
)
from ...definitions import (
    ConditionalSagaStep,
    LocalSagaStep,
    RemoteSagaStep,
    SagaStep,
)
from ..status import (
    SagaStepStatus,
)


class SagaStepExecution(ABC):
    """Saga Step Execution class."""

    def __init__(
        self,
        definition: SagaStep,
        related_services: Optional[set[str]] = None,
        status: SagaStepStatus = SagaStepStatus.Created,
        already_rollback: bool = False,
    ):
        if related_services is None:
            related_services = set()

        self.definition = definition
        self.status = status
        self.already_rollback = already_rollback
        self.related_services = related_services

    @classmethod
    def from_raw(cls, raw: Union[dict[str, Any], SagaStepExecution], **kwargs) -> SagaStepExecution:
        """Build a new instance from a raw representation.

        :param raw: The raw representation of the instance.
        :param kwargs: Additional named arguments.
        :return: A ``SagaStepExecution`` instance.
        """
        if isinstance(raw, cls):
            return raw

        current = raw | kwargs

        if "cls" in current:
            # noinspection PyTypeChecker
            execution_cls: type = import_module(current.pop("cls"))
        else:
            execution_cls = cls

        if not issubclass(execution_cls, cls):
            raise TypeError(f"Given class is not a subclass of {cls}. Obtained: {execution_cls}")

        return execution_cls._from_raw(current)

    @classmethod
    def _from_raw(cls, raw: dict[str, Any]) -> SagaStepExecution:
        raw["definition"] = SagaStep.from_raw(raw["definition"])
        raw["status"] = SagaStepStatus.from_raw(raw["status"])

        if (raw_related_services := raw.get("related_services")) is not None:
            raw["related_services"] = set(raw_related_services)
        return cls(**raw)

    @staticmethod
    def from_definition(step: SagaStep) -> SagaStepExecution:
        """Build a ``SagaStepExecution`` instance from the ``SagaStep`` definition.

        :param step: The ``SagaStep`` definition.
        :return: A new ``SagaStepExecution``.
        """
        from .conditional import (
            ConditionalSagaStepExecution,
        )
        from .local import (
            LocalSagaStepExecution,
        )
        from .remote import (
            RemoteSagaStepExecution,
        )

        if isinstance(step, ConditionalSagaStep):
            return ConditionalSagaStepExecution(step)

        if isinstance(step, LocalSagaStep):
            return LocalSagaStepExecution(step)

        if isinstance(step, RemoteSagaStep):
            return RemoteSagaStepExecution(step)

        raise TypeError(f"Given step is not supported yet. Obtained: {step}")

    @abstractmethod
    async def execute(self, context: SagaContext, *args, **kwargs) -> SagaContext:
        """Execution the step.

        :param context: The execution context to be used during the execution.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: The updated context.
        """

    @abstractmethod
    async def rollback(self, context: SagaContext, *args, **kwargs) -> SagaContext:
        """Revert the executed step.

        :param context: Execution context.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: The updated execution context.
        """

    @property
    def raw(self) -> dict[str, Any]:
        """Compute a raw representation of the instance.

        :return: A ``dict`` instance.
        """
        return {
            "cls": classname(type(self)),
            "definition": self.definition.raw,
            "status": self.status.raw,
            "related_services": list(self.related_services),
            "already_rollback": self.already_rollback,
        }

    def __eq__(self, other: SagaStep) -> bool:
        return type(self) == type(other) and tuple(self) == tuple(other)

    def __iter__(self) -> Iterable:
        yield from (
            self.definition,
            self.status,
            self.related_services,
            self.already_rollback,
        )



from __future__ import (
    annotations,
)

import logging
import warnings
from asyncio import (
    shield,
)
from contextlib import (
    suppress,
)
from typing import (
    Any,
    Iterable,
    Optional,
    Union,
)
from uuid import (
    UUID,
)

from ..context import (
    SagaContext,
)
from ..definitions import (
    Saga,
    SagaDecoratorWrapper,
    SagaStep,
)
from ..exceptions import (
    SagaExecutionAlreadyExecutedException,
    SagaFailedCommitCallbackException,
    SagaFailedExecutionStepException,
    SagaPausedExecutionStepException,
    SagaRollbackExecutionException,
    SagaStepExecutionException,
)
from ..messages import (
    SagaResponse,
)
from .commit import (
    TransactionCommitter,
)
from .status import (
    SagaStatus,
)
from .steps import (
    SagaStepExecution,
)

logger = logging.getLogger(__name__)


class SagaExecution:
    """Saga Execution class."""

    # noinspection PyUnusedLocal
    def __init__(
        self,
        definition: Saga,
        uuid: UUID,
        context: SagaContext,
        status: SagaStatus = SagaStatus.Created,
        steps: list[SagaStepExecution] = None,
        paused_step: SagaStepExecution = None,
        already_rollback: bool = False,
        user: Optional[UUID] = None,
        *args,
        **kwargs,
    ):
        """Initialize a Saga Execution.

        :param definition: The ``Saga`` definition.
        :param uuid: The identifier of the execution.
        :param context: The execution context.
        :param status: The status of the execution.
        :param steps: The executed steps of the execution.
        :param paused_step: The paused step of the execution.
        :param already_rollback: ``True`` if already rollback of ``False`` otherwise.
        :param user: The user that launched the execution.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        """
        definition.validate()  # If not valid, raises an exception.

        if steps is None:
            steps = list()

        self.uuid = uuid
        self.definition = definition
        self.executed_steps = steps
        self.context = context
        self.status = status
        self.already_rollback = already_rollback
        self.paused_step = paused_step
        self.user = user

    @classmethod
    def from_raw(cls, raw: Union[dict[str, Any], SagaExecution], **kwargs) -> SagaExecution:
        """Build a new instance from a raw representation.

        :param raw: The raw representation of the instance.
        :param kwargs: Additional named arguments.
        :return: A ``SagaExecution`` instance.
        """
        if isinstance(raw, cls):
            return raw

        raw = dict(raw)

        current = raw | kwargs
        current["definition"] = Saga.from_raw(current["definition"])
        current["status"] = SagaStatus.from_raw(current["status"])
        current["context"] = SagaContext.from_avro_str(current["context"])
        current["paused_step"] = (
            None if current["paused_step"] is None else SagaStepExecution.from_raw(current["paused_step"])
        )

        if isinstance(current["uuid"], str):
            current["uuid"] = UUID(current["uuid"])
        if isinstance(current["user"], str):
            current["user"] = UUID(current["user"])

        instance = cls(**current)

        executed_steps = (
            SagaStepExecution.from_raw(executed_step, definition=step)
            for step, executed_step in zip(instance.definition.steps, raw.pop("executed_steps"))
        )
        for executed_step in executed_steps:
            instance._add_executed(executed_step)

        return instance

    @classmethod
    def from_saga(cls, definition: Saga, context: Optional[SagaContext] = None, *args, **kwargs) -> SagaExecution:
        """Build a new instance from a ``Saga`` object.

        :param definition: The definition of the saga.
        :param context: Initial saga execution context. If not provided, then a new empty context is created.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: A new ``SagaExecution`` instance.
        """
        warnings.warn(
            "from_saga() method is deprecated by from_definition() and will be removed soon.", DeprecationWarning
        )

        return cls.from_definition(definition, context, *args, **kwargs)

    @classmethod
    def from_definition(
        cls,
        definition: Union[Saga, SagaDecoratorWrapper],
        context: Optional[SagaContext] = None,
        uuid: Optional[UUID] = None,
        *args,
        **kwargs,
    ) -> SagaExecution:
        """Build a new instance from a ``Saga`` object.

        :param definition: The definition of the saga.
        :param context: Initial saga execution context. If not provided, then a new empty context is created.
        :param uuid: The identifier of the execution. If ``None`` is provided, then a new one will be generated.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: A new ``SagaExecution`` instance.
        """
        if isinstance(definition, SagaDecoratorWrapper):
            definition = definition.meta.definition

        if uuid is None:
            from uuid import (
                uuid4,
            )

            uuid = uuid4()

        if context is None:
            context = SagaContext()

        return cls(definition, uuid, context, *args, **kwargs)

    async def execute(
        self, response: Optional[SagaResponse] = None, *args, autocommit: bool = True, **kwargs
    ) -> SagaContext:
        """Execute the ``Saga`` definition.

        :param response: An optional ``SagaResponse`` to be consumed by the immediately next executed step.
        :param args: Additional positional arguments.
        :param autocommit: If ``True`` the commit process is performed automatically, otherwise must be performed
            manually.
        :param kwargs: Additional named arguments.
        :return: A ``SagaContext`` instance.
        """
        if self.status == SagaStatus.Finished:
            raise SagaExecutionAlreadyExecutedException(
                f"The {self.uuid!s} execution cannot be executed because is in {self.status!r} status."
            )
        if self.status == SagaStatus.Errored:
            if response is None:
                raise SagaExecutionAlreadyExecutedException(
                    f"The {self.uuid!s} execution cannot be executed because is in {self.status!r} status."
                )

            logger.info(f"Received 'on_failure': {response!s}")
            return self.context

        self.status = SagaStatus.Running
        if self.paused_step is not None:
            try:
                await self._execute_one(self.paused_step, response=response, autocommit=autocommit, *args, **kwargs)
            finally:
                if self.status != SagaStatus.Paused:
                    self.paused_step = None

        for step in self._pending_steps:
            execution_step = SagaStepExecution.from_definition(step)
            await self._execute_one(execution_step, autocommit=autocommit, *args, **kwargs)

        self.status = SagaStatus.Finished

        if autocommit:
            await self.commit(*args, **kwargs)

        return self.context

    async def _execute_one(self, execution_step: SagaStepExecution, *args, **kwargs) -> None:
        try:
            self.context = await execution_step.execute(
                self.context, execution_uuid=self.uuid, user=self.user, *args, **kwargs
            )
            self._add_executed(execution_step)
        except SagaFailedExecutionStepException as exc:
            await self.rollback(*args, **({"autoreject": kwargs.get("autocommit", True)} | kwargs))
            self.status = SagaStatus.Errored
            raise exc
        except SagaPausedExecutionStepException as exc:
            self.paused_step = execution_step
            self.status = SagaStatus.Paused
            raise exc

    async def rollback(self, *args, autoreject: bool = True, **kwargs) -> None:
        """Revert the executed operation with a compensatory operation.

        :param args: Additional positional arguments.
        :param autoreject: If ``True`` the commit process is performed automatically, otherwise must be performed
            manually.
        :param kwargs: Additional named arguments.
        :return: The updated execution context.
        """

        if self.already_rollback:
            raise SagaRollbackExecutionException("The execution rollback was already performed.")

        raised_exception = False
        for execution_step in reversed(self.executed_steps):
            try:
                self.context = await execution_step.rollback(
                    self.context, user=self.user, execution_uuid=self.uuid, *args, **kwargs
                )
            except SagaStepExecutionException as exc:
                logger.warning(f"There was an exception on {type(execution_step).__name__!r} rollback: {exc!r}")
                raised_exception = True

        if autoreject:
            try:
                await self.reject()
            except SagaFailedCommitCallbackException:
                raised_exception = True

        self.already_rollback = True

        if raised_exception:
            raise SagaRollbackExecutionException("The execution failed during the rollback process.")

    async def commit(self, *args, **kwargs) -> None:
        """Commit the execution transactions.

        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """
        if self.status != SagaStatus.Finished:
            raise ValueError(
                f"Commit can be performed only by {SagaStatus.Finished!r} executions. Obtained: {self.status!r}"
            )

        committer = TransactionCommitter(self.uuid, self.executed_steps, *args, **kwargs)
        try:
            await shield(committer.commit())
        except Exception as exc:  # FIXME: Exception is too broad
            logger.warning(f"There was an exception on {TransactionCommitter.__name__!r} commit: {exc!r}")
            with suppress(SagaRollbackExecutionException):
                await self.rollback(*args, **({"autoreject": False} | kwargs))
            self.status = SagaStatus.Errored
            raise SagaFailedCommitCallbackException(exc)

    async def reject(self, *args, **kwargs) -> None:
        """Reject the execution transactions.

        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """
        committer = TransactionCommitter(self.uuid, self.executed_steps, *args, **kwargs)
        try:
            await shield(committer.reject())
        except Exception as exc:
            logger.warning(f"There was an exception on {TransactionCommitter.__name__!r} rejection: {exc!r}")
            raise SagaFailedCommitCallbackException(exc)

    @property
    def _pending_steps(self) -> list[SagaStep]:
        offset = len(self.executed_steps)
        return self.definition.steps[offset:]

    def _add_executed(self, executed_step: SagaStepExecution) -> None:
        self.executed_steps.append(executed_step)

    @property
    def raw(self) -> dict[str, Any]:
        """Compute a raw representation of the instance.

        :return: A ``dict`` instance.
        """
        return {
            "uuid": str(self.uuid),
            "definition": self.definition.raw,
            "status": self.status.raw,
            "executed_steps": [step.raw for step in self.executed_steps],
            "paused_step": None if self.paused_step is None else self.paused_step.raw,
            "context": self.context.avro_str,
            "already_rollback": self.already_rollback,
            "user": None if self.user is None else str(self.user),
        }

    def __eq__(self, other: SagaStep) -> bool:
        return type(self) == type(other) and tuple(self) == tuple(other)

    def __iter__(self) -> Iterable:
        yield from (
            self.definition,
            self.uuid,
            self.status,
            self.executed_steps,
            self.context,
            self.already_rollback,
            self.user,
        )



import logging
from asyncio import (
    gather,
)
from typing import (
    Optional,
)
from uuid import (
    UUID,
)

from cached_property import (
    cached_property,
)

from minos.common import (
    Inject,
    NotProvidedException,
    PoolFactory,
)
from minos.networks import (
    BrokerClient,
    BrokerClientPool,
    BrokerMessageV1,
    BrokerMessageV1Payload,
    BrokerPublisher,
)

from .steps import (
    ConditionalSagaStepExecution,
    SagaStepExecution,
)

logger = logging.getLogger(__name__)


class TransactionCommitter:
    """Transaction Committer class."""

    # noinspection PyUnusedLocal
    @Inject()
    def __init__(
        self,
        execution_uuid: UUID,
        executed_steps: list[SagaStepExecution],
        broker_publisher: BrokerPublisher,
        broker_pool: Optional[BrokerClientPool] = None,
        pool_factory: Optional[PoolFactory] = None,
        **kwargs,
    ):

        if broker_pool is None and pool_factory is not None:
            broker_pool = pool_factory.get_pool("broker")

        if broker_pool is None:
            raise NotProvidedException(f"A {BrokerClientPool!r} instance is required.")

        self.executed_steps = executed_steps
        self.execution_uuid = execution_uuid

        self.broker_pool = broker_pool
        self.broker_publisher = broker_publisher

    # noinspection PyUnusedCommit,PyMethodOverriding
    async def commit(self, **kwargs) -> None:
        """Commit the transaction.

        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """
        logger.info("committing...")

        if not await self._reserve():
            await self.reject()
            raise ValueError("Some transactions could not be committed.")

        await self._commit()

    async def _reserve(self) -> bool:
        async with self.broker_pool.acquire() as broker:
            futures = (
                broker.send(BrokerMessageV1(f"_Reserve{service_name.title()}Transaction", BrokerMessageV1Payload(uuid)))
                for (uuid, service_name) in self.transactions
            )
            await gather(*futures)

            return await self._get_response(broker, len(self.transactions))

    async def _commit(self) -> None:
        futures = (
            self.broker_publisher.send(
                BrokerMessageV1(f"_Commit{service_name.title()}Transaction", BrokerMessageV1Payload(uuid))
            )
            for (uuid, service_name) in self.transactions
        )
        await gather(*futures)

        # await self._get_response(handler, len(self.transactions))
        logger.info("Successfully committed!")

    async def reject(self) -> None:
        """Reject the transaction.

        :return: This method does not return anything.
        """
        futures = (
            self.broker_publisher.send(
                BrokerMessageV1(f"_Reject{service_name.title()}Transaction", BrokerMessageV1Payload(uuid))
            )
            for (uuid, service_name) in self.transactions
        )
        await gather(*futures)

        logger.info("Successfully rejected!")

    @staticmethod
    async def _get_response(handler: BrokerClient, count: int, **kwargs) -> bool:
        messages = [message async for message in handler.receive_many(count, **kwargs)]
        return all(message.ok for message in messages)

    @cached_property
    def transactions(self) -> list[tuple[UUID, str]]:
        """Get the list of transactions used during the saga execution.

        :return: A list of tuples in which the first value is the identifier of the transaction and the second one is
            the name of the microservice in which the saga was executed.
        """
        transactions = list()
        uniques = set()

        def _fn(uuid: UUID, steps: list[SagaStepExecution]) -> None:
            for step in steps:
                if isinstance(step, ConditionalSagaStepExecution):
                    inner = step.inner
                    if inner is not None:
                        _fn(inner.uuid, inner.executed_steps)
                else:
                    for service_name in step.related_services:
                        pair = (uuid, service_name)
                        if pair not in uniques:
                            transactions.append(pair)
                            uniques.add(pair)

        _fn(self.execution_uuid, self.executed_steps)
        transactions.sort()
        return transactions



"""The SAGA pattern of the Minos Framework."""

__author__ = "Minos Framework Devs"
__email__ = "hey@minos.run"
__version__ = "0.8.0.dev3"

from .context import (
    SagaContext,
)
from .definitions import (
    ConditionalSagaStep,
    ConditionalSagaStepDecoratorMeta,
    ConditionalSagaStepDecoratorWrapper,
    ElseThenAlternative,
    ElseThenAlternativeDecoratorMeta,
    ElseThenAlternativeDecoratorWrapper,
    IfThenAlternative,
    IfThenAlternativeDecoratorMeta,
    IfThenAlternativeDecoratorWrapper,
    LocalSagaStep,
    LocalSagaStepDecoratorMeta,
    LocalSagaStepDecoratorWrapper,
    RemoteSagaStep,
    RemoteSagaStepDecoratorMeta,
    RemoteSagaStepDecoratorWrapper,
    Saga,
    SagaDecoratorMeta,
    SagaDecoratorWrapper,
    SagaOperation,
    SagaOperationDecorator,
    SagaStep,
    SagaStepDecoratorMeta,
    SagaStepDecoratorWrapper,
)
from .exceptions import (
    AlreadyCommittedException,
    AlreadyOnSagaException,
    EmptySagaException,
    EmptySagaStepException,
    MultipleElseThenException,
    MultipleOnErrorException,
    MultipleOnExecuteException,
    MultipleOnFailureException,
    MultipleOnSuccessException,
    OrderPrecedenceException,
    SagaException,
    SagaExecutionAlreadyExecutedException,
    SagaExecutionException,
    SagaExecutionNotFoundException,
    SagaFailedCommitCallbackException,
    SagaFailedExecutionException,
    SagaFailedExecutionStepException,
    SagaNotCommittedException,
    SagaNotDefinedException,
    SagaPausedExecutionStepException,
    SagaResponseException,
    SagaRollbackExecutionException,
    SagaRollbackExecutionStepException,
    SagaStepException,
    SagaStepExecutionException,
    UndefinedOnExecuteException,
)
from .executions import (
    ConditionalSagaStepExecution,
    DatabaseSagaExecutionRepository,
    Executor,
    LocalExecutor,
    LocalSagaStepExecution,
    RemoteSagaStepExecution,
    RequestExecutor,
    ResponseExecutor,
    SagaExecution,
    SagaExecutionDatabaseOperationFactory,
    SagaExecutionRepository,
    SagaRunner,
    SagaStatus,
    SagaStepExecution,
    SagaStepStatus,
    TransactionCommitter,
)
from .managers import (
    SagaManager,
)
from .messages import (
    SagaRequest,
    SagaResponse,
    SagaResponseStatus,
)
from .middleware import (
    transactional_command,
)
from .services import (
    SagaService,
)
from .utils import (
    get_service_name,
)



"""Conditional Step Definitions module."""

from __future__ import (
    annotations,
)

from collections.abc import (
    Callable,
    Iterable,
)
from inspect import (
    getmembers,
)
from operator import (
    attrgetter,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Protocol,
    TypeVar,
    Union,
    runtime_checkable,
)

from cached_property import (
    cached_property,
)

from minos.common import (
    classname,
)

from ...exceptions import (
    EmptySagaStepException,
    MultipleElseThenException,
    OrderPrecedenceException,
)
from ..operations import (
    SagaOperation,
)
from ..types import (
    ConditionCallback,
)
from .abc import (
    SagaStep,
    SagaStepDecoratorMeta,
    SagaStepDecoratorWrapper,
)

if TYPE_CHECKING:
    from ..saga import (
        Saga,
    )


@runtime_checkable
class ConditionalSagaStepDecoratorWrapper(SagaStepDecoratorWrapper, Protocol):
    """Conditional Saga Step Decorator Wrapper class."""

    meta: ConditionalSagaStepDecoratorMeta


class ConditionalSagaStepDecoratorMeta(SagaStepDecoratorMeta):
    """Conditional Saga Step Decorator Meta class."""

    _definition: ConditionalSagaStep

    @cached_property
    def definition(self):
        """Get the step definition.

        :return: A ``SagaStep`` instance.
        """
        if_then_alternatives = getmembers(
            self._inner,
            predicate=(
                lambda x: isinstance(x, IfThenAlternativeDecoratorWrapper)
                and isinstance(x.meta, IfThenAlternativeDecoratorMeta)
            ),
        )
        if_then_alternatives = list(map(lambda member: member[1].meta.alternative, if_then_alternatives))
        for alternative in if_then_alternatives:
            if alternative.order is None:
                raise OrderPrecedenceException(f"The {alternative!r} alternative does not have 'order' value.")
        if_then_alternatives.sort(key=attrgetter("order"))

        for alternative in if_then_alternatives:
            self._definition.if_then(alternative)

        else_then_alternatives = getmembers(
            self._inner,
            predicate=(
                lambda x: isinstance(x, ElseThenAlternativeDecoratorWrapper)
                and isinstance(x.meta, ElseThenAlternativeDecoratorMeta)
            ),
        )
        else_then_alternatives = list(map(lambda member: member[1].meta.alternative, else_then_alternatives))
        for alternative in else_then_alternatives:
            self._definition.else_then(alternative)

        return self._definition


TP = TypeVar("TP", bound=type)


class ConditionalSagaStep(SagaStep):
    """Conditional Saga Step class."""

    if_then_alternatives: list[IfThenAlternative]

    else_then_alternative: ElseThenAlternative

    def __init__(
        self,
        if_then: Optional[Union[IfThenAlternative, Iterable[IfThenAlternative]]] = None,
        else_then: Optional[ElseThenAlternative] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        if isinstance(if_then, IfThenAlternative):
            if_then = [if_then]

        if if_then is None:
            if_then = list()

        self.if_then_alternatives = list(if_then)
        self.else_then_alternative = else_then

    @classmethod
    def _from_raw(cls, raw: dict[str, Any], **kwargs) -> ConditionalSagaStep:
        current = raw | kwargs

        if_then = current.get("if_then", None)
        if if_then is not None:
            current["if_then"] = [IfThenAlternative.from_raw(alternative) for alternative in if_then]

        else_then = current.get("else_then", None)
        if else_then is not None:
            current["else_then"] = ElseThenAlternative.from_raw(else_then)

        return cls(**current)

    def __call__(self, type_: TP) -> Union[TP, ConditionalSagaStepDecoratorWrapper]:
        """Decorate the given type.

        :param type_: The type to be decorated.
        :return: The decorated type.
        """
        meta = ConditionalSagaStepDecoratorMeta(type_, self)
        type_.meta = meta
        return type_

    def if_then(
        self, alternative: Union[IfThenAlternative, ConditionCallback], saga: Optional[Saga] = None
    ) -> ConditionalSagaStep:
        """Add a new ``IfThenAlternative`` based on a condition and a saga.

        :param alternative: The condition that must be satisfied to execute the alternative.
        :param saga: The saga to be executed if the condition is satisfied.
        :return: This method returns the same instance that is called.
        """
        if not isinstance(alternative, IfThenAlternative):
            alternative = IfThenAlternative(saga, alternative)

        if alternative.order is None:
            if self.if_then_alternatives:
                alternative.order = self.if_then_alternatives[-1].order + 1
            else:
                alternative.order = 1

        if self.if_then_alternatives and alternative.order <= self.if_then_alternatives[-1].order:
            raise OrderPrecedenceException(
                f"Unsatisfied precedence constraints. Previous: {self.if_then_alternatives[-1].order} "
                f"Current: {alternative.order} "
            )

        self.if_then_alternatives.append(alternative)
        return self

    def else_then(self, alternative: Union[ElseThenAlternative, Saga]) -> ConditionalSagaStep:
        """Set the ``ElseThenAlternative`` with the given saga.

        :param alternative: The saga to be executed if not any condition is met.
        :return: This method returns the same instance that is called.
        """

        if self.else_then_alternative is not None:
            raise MultipleElseThenException()

        if not isinstance(alternative, ElseThenAlternative):
            alternative = ElseThenAlternative(alternative)

        self.else_then_alternative = alternative
        return self

    def validate(self) -> None:
        """Check if the step is valid.

        :return: This method does not return anything, but raises an exception if the step is not valid.
        """

        if not len(self.if_then_alternatives) and self.else_then_alternative is None:
            raise EmptySagaStepException()

        for alternative in self.if_then_alternatives:
            alternative.validate()

        if self.else_then_alternative is not None:
            self.else_then_alternative.validate()

    @property
    def raw(self) -> dict[str, Any]:
        """Get the raw representation of the step.

        :return: A ``dict`` instance.
        """

        return {
            "cls": classname(type(self)),
            "order": self.order,
            "if_then": [alternative.raw for alternative in self.if_then_alternatives],
            "else_then": None if self.else_then_alternative is None else self.else_then_alternative.raw,
        }

    def __iter__(self) -> Iterable:
        yield from (
            self.order,
            self.if_then_alternatives,
            self.else_then_alternative,
        )


@runtime_checkable
class IfThenAlternativeDecoratorWrapper(Protocol):
    """If Then Alternative Decorator Wrapper class."""

    meta: IfThenAlternativeDecoratorMeta
    __call__: ConditionCallback


class IfThenAlternativeDecoratorMeta:
    """If Then Alternative Decorator Meta class."""

    def __init__(self, inner: ConditionCallback, alternative: IfThenAlternative):
        self._inner = inner
        self._alternative = alternative

    @cached_property
    def alternative(self) -> IfThenAlternative:
        """Get the if-then alternative.

        :return: A ``IfThenAlternative`` instance.
        """
        self._alternative.condition = SagaOperation(self._inner)
        return self._alternative


class IfThenAlternative:
    """If Then Alternative class."""

    def __init__(
        self,
        saga: Saga,
        condition: Optional[Union[ConditionCallback, SagaOperation[ConditionCallback]]] = None,
        order: Optional[int] = None,
    ):
        if not isinstance(condition, SagaOperation):
            condition = SagaOperation(condition)

        self.condition = condition
        self.saga = saga
        self.order = order

    @classmethod
    def from_raw(cls, raw: Union[dict[str, Any], IfThenAlternative], **kwargs) -> IfThenAlternative:
        """Build a new instance from a raw representation.

        :param raw: The raw representation.
        :param kwargs: Additional named arguments.
        :return: A new ``IfThenAlternative`` instance.
        """

        from ..saga import (
            Saga,
        )

        if isinstance(raw, cls):
            return raw
        current = raw | kwargs

        current["condition"] = SagaOperation.from_raw(current["condition"])
        current["saga"] = Saga.from_raw(current["saga"])

        return cls(**current)

    def __call__(self, func: ConditionCallback) -> IfThenAlternativeDecoratorWrapper:
        meta = IfThenAlternativeDecoratorMeta(func, self)
        func.meta = meta
        return func

    def validate(self) -> None:
        """Check if the alternative is valid.

        :return: This method does not return anything, but raises an exception if the alternative is not valid.
        """

        return self.saga.validate()

    @property
    def raw(self) -> dict[str, Any]:
        """Get the raw representation of the alternative.

        :return: A ``dict`` value.
        """

        return {
            "order": self.order,
            "condition": self.condition.raw,
            "saga": self.saga.raw,
        }

    def __repr__(self) -> str:
        return f"{type(self).__name__}{tuple(self)}"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and tuple(self) == tuple(other)

    def __iter__(self):
        yield from (
            self.condition,
            self.saga,
        )


@runtime_checkable
class ElseThenAlternativeDecoratorWrapper(Protocol):
    """Else Then Alternative Decorator Wrapper class."""

    meta: ElseThenAlternativeDecoratorMeta
    __call__: Callable


class ElseThenAlternativeDecoratorMeta:
    """Else Then Alternative Decorator Meta class."""

    def __init__(self, inner: Callable, alternative: ElseThenAlternative):
        self._inner = inner
        self._alternative = alternative

    @cached_property
    def alternative(self) -> ElseThenAlternative:
        """Get the else-then alternative.

        :return: A ``ElseThenAlternative`` instance.
        """
        return self._alternative


class ElseThenAlternative:
    """Else Then Alternative class."""

    def __init__(self, saga: Saga):
        self.saga = saga

    @classmethod
    def from_raw(cls, raw: Union[dict[str, Any], ElseThenAlternative], **kwargs) -> ElseThenAlternative:
        """Build a new instance from a raw representation.

        :param raw: The raw representation.
        :param kwargs: Additional named arguments.
        :return: A new ``ElseThenAlternative`` instance.
        """

        from ..saga import (
            Saga,
        )

        if isinstance(raw, cls):
            return raw
        current = raw | kwargs

        current["saga"] = Saga.from_raw(current["saga"])

        return cls(**current)

    def __call__(self, func: Callable) -> ElseThenAlternativeDecoratorWrapper:
        meta = ElseThenAlternativeDecoratorMeta(func, self)
        func.meta = meta
        return func

    def validate(self) -> None:
        """Check if the alternative is valid.

        :return: This method does not return anything, but raises an exception if the alternative is not valid.
        """

        return self.saga.validate()

    @property
    def raw(self) -> dict[str, Any]:
        """Get the raw representation of the alternative.

        :return: A ``dict`` value.
        """

        return {
            "saga": self.saga.raw,
        }

    def __repr__(self) -> str:
        return f"{type(self).__name__}{tuple(self)}"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and tuple(self) == tuple(other)

    def __iter__(self):
        yield from (self.saga,)



from __future__ import (
    annotations,
)

from contextlib import (
    suppress,
)
from itertools import (
    chain,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Optional,
)
from uuid import (
    UUID,
)

from ...context import (
    SagaContext,
)
from ...definitions import (
    ConditionalSagaStep,
)
from ...exceptions import (
    SagaExecutionAlreadyExecutedException,
    SagaFailedExecutionStepException,
    SagaPausedExecutionStepException,
    SagaRollbackExecutionStepException,
)
from ...utils import (
    get_service_name,
)
from ..executors import (
    Executor,
)
from ..status import (
    SagaStepStatus,
)
from .abc import (
    SagaStepExecution,
)

if TYPE_CHECKING:
    from ..saga import (
        SagaExecution,
    )


class ConditionalSagaStepExecution(SagaStepExecution):
    """Conditional Saga Step Execution class."""

    inner: Optional[SagaExecution]
    definition: ConditionalSagaStep

    def __init__(self, *args, inner: Optional[SagaExecution] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.inner = inner

    @classmethod
    def _from_raw(cls, raw: dict[str, Any]) -> SagaStepExecution:
        from ...executions import (
            SagaExecution,
        )

        raw["inner"] = SagaExecution.from_raw(raw["inner"])

        return super()._from_raw(raw)

    async def execute(self, context: SagaContext, *args, **kwargs) -> SagaContext:
        """Execution the step.

        :param context: The execution context to be used during the execution.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: The updated context.
        """

        if self.status == SagaStepStatus.Created:
            self.inner = await self._create_inner(context, *args, **kwargs)

        if self.inner is not None:
            context = await self._execute_inner(context, *args, **kwargs)

        self.status = SagaStepStatus.Finished

        return context

    async def _create_inner(
        self, context: SagaContext, user: Optional[UUID] = None, execution_uuid: Optional[UUID] = None, *args, **kwargs
    ) -> Optional[SagaExecution]:
        from ...executions import (
            SagaExecution,
        )

        self.status = SagaStepStatus.RunningOnExecute

        executor = Executor(execution_uuid=execution_uuid)
        self.related_services.add(get_service_name())

        definition = None
        for alternative in self.definition.if_then_alternatives:
            if await executor.exec(alternative.condition, context):
                definition = alternative.saga
                break

        if definition is None and self.definition.else_then_alternative is not None:
            definition = self.definition.else_then_alternative.saga

        execution = None
        if definition is not None:
            execution = SagaExecution.from_definition(
                definition, context=context, user=user, uuid=execution_uuid, *args, **kwargs
            )

        self.status = SagaStepStatus.FinishedOnExecute

        return execution

    async def _execute_inner(
        self, context: SagaContext, user: Optional[UUID] = None, execution_uuid: Optional[UUID] = None, *args, **kwargs
    ) -> SagaContext:
        execution = self.inner
        if execution_uuid is not None:
            execution.uuid = execution_uuid
        if user is not None:
            execution.user = user
        execution.context = context

        self.status = SagaStepStatus.RunningOnSuccess

        try:
            with suppress(SagaExecutionAlreadyExecutedException):
                await self.inner.execute(*args, **(kwargs | {"autocommit": False}))
        except SagaPausedExecutionStepException as exc:
            self.status = SagaStepStatus.PausedByOnExecute
            raise exc
        except SagaFailedExecutionStepException as exc:
            self.status = SagaStepStatus.ErroredByOnExecute
            raise exc

        self.status = SagaStepStatus.FinishedOnSuccess

        return execution.context

    async def rollback(self, context: SagaContext, *args, **kwargs) -> SagaContext:
        """Revert the executed step.

        :param context: Execution context.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: The updated execution context.
        """
        if self.status == SagaStepStatus.Created:
            raise SagaRollbackExecutionStepException("There is nothing to rollback.")

        if self.already_rollback:
            raise SagaRollbackExecutionStepException("The step was already rollbacked.")

        if self.inner is not None:
            context = await self._rollback_inner(context, *args, **kwargs)

        self.already_rollback = True
        return context

    async def _rollback_inner(
        self, context: SagaContext, user: Optional[UUID] = None, execution_uuid: Optional[UUID] = None, *args, **kwargs
    ) -> SagaContext:
        execution = self.inner
        if execution_uuid is not None:
            execution.uuid = execution_uuid
        if user is not None:
            execution.user = user
        execution.context = context

        await execution.rollback(*args, **(kwargs | {"autoreject": False}))

        return execution.context

    @property
    def raw(self) -> dict[str, Any]:
        """Compute a raw representation of the instance.

        :return: A ``dict`` instance.
        """
        return super().raw | {"inner": None if self.inner is None else self.inner.raw}

    def __iter__(self) -> Iterable:
        yield from chain(super().__iter__(), (self.inner,))



from __future__ import (
    annotations,
)

from typing import (
    Optional,
)

from ...context import (
    SagaContext,
)
from ...definitions import (
    RemoteSagaStep,
)
from ...exceptions import (
    SagaFailedExecutionStepException,
    SagaPausedExecutionStepException,
    SagaResponseException,
    SagaRollbackExecutionStepException,
)
from ...messages import (
    SagaResponse,
    SagaResponseStatus,
)
from ...utils import (
    get_service_name,
)
from ..executors import (
    RequestExecutor,
    ResponseExecutor,
)
from ..status import (
    SagaStepStatus,
)
from .abc import (
    SagaStepExecution,
)


class RemoteSagaStepExecution(SagaStepExecution):
    """Saga Execution Step class."""

    definition: RemoteSagaStep

    async def execute(
        self, context: SagaContext, response: Optional[SagaResponse] = None, *args, **kwargs
    ) -> SagaContext:
        """Execution the remote step.

        :param context: The execution context to be used during the execution.
        :param response: An optional command response instance (to be consumed by the on_success method).
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: The updated context.
        """

        await self._execute_on_execute(context, *args, **kwargs)

        if response is None:
            self.status = SagaStepStatus.PausedByOnExecute
            raise SagaPausedExecutionStepException()

        self.related_services |= response.related_services

        if response.status == SagaResponseStatus.SYSTEM_ERROR:
            self.status = SagaStepStatus.ErroredByOnExecute
            exc = SagaResponseException(f"Failed with {response.status!s} status: {await response.content()!s}")
            raise SagaFailedExecutionStepException(exc)

        if response.status == SagaResponseStatus.SUCCESS:
            context = await self._execute_on_success(context, response, *args, **kwargs)
        else:
            context = await self._execute_on_error(context, response, *args, **kwargs)

        self.status = SagaStepStatus.Finished
        return context

    async def _execute_on_execute(self, context: SagaContext, *args, **kwargs) -> None:
        if self.status != SagaStepStatus.Created:
            return

        self.status = SagaStepStatus.RunningOnExecute
        executor = RequestExecutor(*args, **kwargs)
        self.related_services.add(get_service_name())
        try:
            await executor.exec(self.definition.on_execute_operation, context)
        except SagaFailedExecutionStepException as exc:
            self.status = SagaStepStatus.ErroredOnExecute
            raise exc
        self.status = SagaStepStatus.FinishedOnExecute

    async def _execute_on_success(self, context: SagaContext, response: SagaResponse, *args, **kwargs) -> SagaContext:
        self.status = SagaStepStatus.RunningOnSuccess
        executor = ResponseExecutor(*args, **kwargs)
        self.related_services.add(get_service_name())

        try:
            context = await executor.exec(self.definition.on_success_operation, context, response)
        except SagaFailedExecutionStepException as exc:
            self.status = SagaStepStatus.ErroredOnSuccess
            await self.rollback(context, *args, **kwargs)
            raise exc

        self.status = SagaStepStatus.FinishedOnSuccess

        return context

    async def _execute_on_error(self, context: SagaContext, response: SagaResponse, *args, **kwargs) -> SagaContext:
        self.status = SagaStepStatus.RunningOnError
        executor = ResponseExecutor(*args, **kwargs)
        self.related_services.add(get_service_name())

        try:
            context = await executor.exec(self.definition.on_error_operation, context, response)
        except SagaFailedExecutionStepException as exc:
            self.status = SagaStepStatus.ErroredOnError
            await self.rollback(context, *args, **kwargs)
            raise exc

        self.status = SagaStepStatus.FinishedOnError

        return context

    async def rollback(self, context: SagaContext, *args, **kwargs) -> SagaContext:
        """Revert the executed remote step.

        :param context: Execution context.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: The updated execution context.
        """
        if self.status == SagaStepStatus.Created:
            raise SagaRollbackExecutionStepException("There is nothing to rollback.")

        if self.already_rollback:
            raise SagaRollbackExecutionStepException("The step was already rollbacked.")

        executor = RequestExecutor(*args, **kwargs)
        await executor.exec(self.definition.on_failure_operation, context)

        self.already_rollback = True
        return context



from __future__ import (
    annotations,
)

from minos.common import (
    Config,
    Inject,
    NotProvidedException,
)


@Inject()
def get_service_name(config: Config) -> str:
    """Get the service name."""
    if config is None:
        raise NotProvidedException("The config object must be provided.")
    return config.get_name()



"""Managers module."""

from __future__ import (
    annotations,
)

from typing import (
    Optional,
    Union,
)
from uuid import (
    UUID,
)

from minos.common import (
    Config,
    Inject,
    Injectable,
    SetupMixin,
)

from .context import (
    SagaContext,
)
from .definitions import (
    Saga,
    SagaDecoratorWrapper,
)
from .executions import (
    DatabaseSagaExecutionRepository,
    SagaExecution,
    SagaExecutionRepository,
    SagaRunner,
)
from .messages import (
    SagaResponse,
)


@Injectable("saga_manager")
class SagaManager(SetupMixin):
    """Saga Manager implementation class.

    The purpose of this class is to manage the running process for new or paused``SagaExecution`` instances.
    """

    @Inject()
    def __init__(
        self,
        storage: SagaExecutionRepository,
        runner: SagaRunner,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.storage = storage
        self.runner = runner

    @classmethod
    def _from_config(cls, config: Config, **kwargs) -> SagaManager:
        """Build an instance from config.

        :param args: Additional positional arguments.
        :param config: Config instance.
        :param kwargs: Additional named arguments.
        :return: A new ``SagaManager`` instance.
        """
        if "storage" not in kwargs:
            kwargs["storage"] = DatabaseSagaExecutionRepository.from_config(config, **kwargs)
        if "runner" not in kwargs:
            kwargs["runner"] = SagaRunner.from_config(config, **kwargs)
        # noinspection PyTypeChecker
        return super()._from_config(config, **kwargs)

    async def _setup(self) -> None:
        await super()._setup()
        await self.storage.setup()
        await self.runner.setup()

    async def _destroy(self) -> None:
        await self.runner.destroy()
        await self.storage.destroy()
        await super()._destroy()

    async def run(
        self,
        definition: Optional[Union[Saga, SagaDecoratorWrapper]] = None,
        context: Optional[SagaContext] = None,
        *,
        response: Optional[SagaResponse] = None,
        user: Optional[UUID] = None,
        autocommit: bool = True,
        pause_on_disk: bool = False,
        raise_on_error: bool = True,
        return_execution: bool = True,
        **kwargs,
    ) -> Union[UUID, SagaExecution]:
        """Perform a run of a ``Saga``.

        The run can be a new one (if a name is provided) or continue execution a previous one (if a reply is provided).

        :param definition: Saga definition to be executed.
        :param context: Initial context to be used during the execution. (Only used for new executions)
        :param response: The reply that relaunches a saga execution.
        :param user: The user identifier to be injected on remote steps.
        :param autocommit: If ``True`` the transactions are committed/rejected automatically. Otherwise, the ``commit``
            or ``reject`` must be called manually.
        :param pause_on_disk: If ``True`` the pauses until remote steps' responses are paused on disk (background,
            non-blocking the execution). Otherwise, the pauses are waited on memory (online, blocking the execution)
        :param raise_on_error: If ``True`` exceptions are raised on error. Otherwise, the execution is returned normally
            but with ``Errored`` status.
        :param return_execution: If ``True`` the ``SagaExecution`` instance is returned. Otherwise, only the
            identifier (``UUID``) is returned.
        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """
        return await self.runner.run(
            definition=definition,
            context=context,
            response=response,
            user=user,
            autocommit=autocommit,
            pause_on_disk=pause_on_disk,
            raise_on_error=raise_on_error,
            return_execution=return_execution,
            **kwargs,
        )

    async def get(self, uuid: UUID) -> SagaExecution:
        """Get a saga execution by identifier.

        :param uuid: The identifier of the execution..
        :return: A ``SagaExecution`` instance.
        """
        return await self.storage.load(uuid)



"""Base Step Definitions module."""

from __future__ import (
    annotations,
)

import warnings
from abc import (
    ABC,
    abstractmethod,
)
from collections.abc import (
    Callable,
    Iterable,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Protocol,
    Union,
    runtime_checkable,
)

from minos.common import (
    import_module,
)

from ...exceptions import (
    SagaNotDefinedException,
)

if TYPE_CHECKING:
    from ..saga import (
        Saga,
    )
    from .conditional import (
        ConditionalSagaStep,
    )
    from .local import (
        LocalSagaStep,
    )
    from .remote import (
        RemoteSagaStep,
    )


@runtime_checkable
class SagaStepDecoratorWrapper(Protocol):
    """Saga Step Decorator Wrapper class."""

    meta: SagaStepDecoratorMeta


class SagaStepDecoratorMeta:
    """Saga Step Decorator Meta class."""

    _inner: Any
    _definition: SagaStep

    def __init__(self, inner: Any, definition: SagaStep):
        self._inner = inner
        self._definition = definition

    @property
    @abstractmethod
    def definition(self) -> SagaStep:
        """Get the step definition.

        :return: A ``SagaStep`` instance.
        """


class SagaStep(ABC):
    """Saga step class."""

    def __init__(self, saga: Optional[Saga] = None, order: Optional[int] = None, **kwargs):
        self.saga = saga
        self.order = order

    @classmethod
    def from_raw(cls, raw: Union[dict[str, Any], SagaStep], **kwargs) -> SagaStep:
        """Build a new instance from raw.

        :param raw: A raw representation.
        :param kwargs: Additional named arguments.
        :return: A ``SagaStep`` instance.
        """
        if isinstance(raw, cls):
            return raw

        raw = raw.copy()

        if "cls" in raw:
            # noinspection PyTypeChecker
            step_cls: type = import_module(raw.pop("cls"))
        else:
            step_cls = cls

        if not issubclass(step_cls, cls):
            raise TypeError(f"Given class is not a subclass of {cls}. Obtained: {step_cls}")

        return step_cls._from_raw(raw | kwargs)

    @classmethod
    @abstractmethod
    def _from_raw(cls, raw: dict[str, Any]) -> SagaStep:
        raise NotImplementedError

    @abstractmethod
    def __call__(self, func: Callable) -> SagaStepDecoratorWrapper:
        """Decorate the given function.

        :param func: The function to be decorated.
        :return: The decorated function.
        """

    def conditional_step(self, *args, **kwargs) -> ConditionalSagaStep:
        """Create a new conditional step in the ``Saga``.

        :param args: Additional positional parameters.
        :param kwargs: Additional named parameters.
        :return: A new ``SagaStep`` instance.
        """
        self.validate()
        if self.saga is None:
            raise SagaNotDefinedException()
        return self.saga.conditional_step(*args, **kwargs)

    def local_step(self, *args, **kwargs) -> LocalSagaStep:
        """Create a new local step in the ``Saga``.

        :param args: Additional positional parameters.
        :param kwargs: Additional named parameters.
        :return: A new ``SagaStep`` instance.
        """
        self.validate()
        if self.saga is None:
            raise SagaNotDefinedException()
        return self.saga.local_step(*args, **kwargs)

    def step(self, *args, **kwargs) -> SagaStep:
        """Create a new step in the ``Saga``.

        :param args: Additional positional parameters.
        :param kwargs: Additional named parameters.
        :return: A new ``SagaStep`` instance.
        """
        warnings.warn("step() method is deprecated by remote_step() and will be removed soon.", DeprecationWarning)
        return self.remote_step(*args, **kwargs)

    def remote_step(self, *args, **kwargs) -> RemoteSagaStep:
        """Create a new remote step in the ``Saga``.

        :param args: Additional positional parameters.
        :param kwargs: Additional named parameters.
        :return: A new ``SagaStep`` instance.
        """
        self.validate()
        if self.saga is None:
            raise SagaNotDefinedException()
        return self.saga.remote_step(*args, **kwargs)

    def commit(self, *args, **kwargs) -> Saga:
        """Commit the current ``SagaStep`` on the ``Saga``.

        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: A ``Saga`` instance.
        """
        self.validate()
        if self.saga is None:
            raise SagaNotDefinedException()
        return self.saga.commit(*args, **kwargs)

    @abstractmethod
    def validate(self) -> None:
        """Check if the step is valid.

        :return: This method does not return anything, but raises an exception if the step is not valid.
        """

    @property
    @abstractmethod
    def raw(self) -> dict[str, Any]:
        """Generate a raw representation of the instance.

        :return: A ``dict`` instance.
        """

    def __hash__(self):
        return hash(tuple(self))

    def __eq__(self, other: SagaStep) -> bool:
        return type(self) == type(other) and tuple(self) == tuple(other)

    def __repr__(self) -> str:
        return f"{type(self).__name__}{tuple(self)}"

    @abstractmethod
    def __iter__(self) -> Iterable:
        """Iterate over the ``SagaStep`` attributes.

        :return: An iterable of attributes.
        """



from .abc import (
    SagaStep,
    SagaStepDecoratorMeta,
    SagaStepDecoratorWrapper,
)
from .conditional import (
    ConditionalSagaStep,
    ConditionalSagaStepDecoratorMeta,
    ConditionalSagaStepDecoratorWrapper,
    ElseThenAlternative,
    ElseThenAlternativeDecoratorMeta,
    ElseThenAlternativeDecoratorWrapper,
    IfThenAlternative,
    IfThenAlternativeDecoratorMeta,
    IfThenAlternativeDecoratorWrapper,
)
from .local import (
    LocalSagaStep,
    LocalSagaStepDecoratorMeta,
    LocalSagaStepDecoratorWrapper,
)
from .remote import (
    RemoteSagaStep,
    RemoteSagaStepDecoratorMeta,
    RemoteSagaStepDecoratorWrapper,
)



from __future__ import (
    annotations,
)

from typing import (
    Optional,
)
from uuid import (
    UUID,
)

from minos.common import (
    Config,
    Inject,
    NotProvidedException,
)
from minos.networks import (
    REQUEST_HEADERS_CONTEXT_VAR,
    REQUEST_REPLY_TOPIC_CONTEXT_VAR,
    BrokerMessageV1,
    BrokerMessageV1Payload,
    BrokerPublisher,
)

from ...context import (
    SagaContext,
)
from ...definitions import (
    RequestCallBack,
    SagaOperation,
)
from ...exceptions import (
    ExecutorException,
    SagaFailedExecutionStepException,
)
from ...messages import (
    SagaRequest,
)
from .abc import (
    Executor,
)


class RequestExecutor(Executor):
    """Request Executor class.

    This class has the responsibility to publish command on the corresponding broker's queue.
    """

    @Inject()
    def __init__(self, *args, user: Optional[UUID], broker_publisher: BrokerPublisher, **kwargs):
        super().__init__(*args, **kwargs)

        self.user = user

        if broker_publisher is None:
            raise NotProvidedException("A broker instance is required.")

        self.broker_publisher = broker_publisher

    # noinspection PyMethodOverriding
    async def exec(self, operation: Optional[SagaOperation[RequestCallBack]], context: SagaContext) -> SagaContext:
        """Exec method, that perform the publishing logic run an pre-callback function to generate the command contents.

        :param operation: Operation to be executed.
        :param context: Execution context.
        :return: A saga context instance.
        """
        if operation is None:
            return context

        try:
            context = SagaContext(**context)  # Needed to avoid mutability issues.
            request = await super().exec(operation, context)
            await self._publish(request)
        except ExecutorException as exc:
            raise SagaFailedExecutionStepException(exc.exception)
        return context

    # noinspection PyMethodOverriding
    async def _publish(self, request: SagaRequest) -> None:
        reply_topic = REQUEST_REPLY_TOPIC_CONTEXT_VAR.get()
        if reply_topic is None:
            reply_topic = self._get_default_reply_topic()

        headers = (REQUEST_HEADERS_CONTEXT_VAR.get() or dict()).copy()
        headers["saga"] = str(self.execution_uuid)
        if self.user is not None:
            headers["user"] = str(self.user)
        if headers.get("transactions"):
            headers["transactions"] += f",{self.execution_uuid!s}"
        else:
            headers["transactions"] = f"{self.execution_uuid!s}"

        try:
            message = BrokerMessageV1(
                topic=request.target,
                payload=BrokerMessageV1Payload(content=await request.content(), headers=headers),
                reply_topic=reply_topic,
            )

            await self.broker_publisher.send(message)
        except Exception as exc:
            raise ExecutorException(exc)

    @Inject()
    def _get_default_reply_topic(self, config: Config) -> str:
        return f"{config.get_name()}Reply"



from .abc import (
    SagaStepExecution,
)
from .conditional import (
    ConditionalSagaStepExecution,
)
from .local import (
    LocalSagaStepExecution,
)
from .remote import (
    RemoteSagaStepExecution,
)



from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    Optional,
)
from uuid import (
    UUID,
)

from minos.common import (
    DatabaseOperation,
    DatabaseOperationFactory,
)


class SagaExecutionDatabaseOperationFactory(DatabaseOperationFactory, ABC):
    """Saga Execution Database Operation Factory class."""

    @abstractmethod
    def build_create(self) -> DatabaseOperation:
        """Build the database operation to create the delta table.

        :return: A ``DatabaseOperation`` instance.s
        """

    @abstractmethod
    def build_store(
        self,
        uuid: UUID,
        definition: dict[str, Any],
        status: str,
        executed_steps: list[dict[str, Any]],
        paused_step: Optional[dict[str, Any]],
        context: str,
        already_rollback: bool,
        user: Optional[UUID],
        **kwargs
    ) -> DatabaseOperation:
        """Build the database operation to store a saga execution.

        :param uuid: The identifier of the saga execution.
        :param definition: The ``Saga`` definition.
        :param context: The execution context.
        :param status: The status of the execution.
        :param executed_steps: The executed steps of the execution.
        :param paused_step: The paused step of the execution.
        :param already_rollback: ``True`` if already rollback of ``False`` otherwise.
        :param user: The user that launched the execution.
        :param kwargs: The attributes of the saga execution.
        :return: A ``DatabaseOperation`` instance.
        """
        raise NotImplementedError

    @abstractmethod
    def build_load(self, uuid: UUID) -> DatabaseOperation:
        """Build the database operation to load a saga execution.

        :param uuid: The identifier of the saga execution.
        :return: A ``DatabaseOperation`` instance.
        """
        raise NotImplementedError

    @abstractmethod
    def build_delete(self, uuid: UUID) -> DatabaseOperation:
        """Build the database operation to delete a saga execution.

        :param uuid: The identifier of the saga execution.
        :return: A ``DatabaseOperation`` instance.
        """
        raise NotImplementedError



"""Local Step Definitions module."""

from __future__ import (
    annotations,
)

from collections.abc import (
    Iterable,
)
from functools import (
    partial,
)
from typing import (
    Any,
    Optional,
    Protocol,
    Union,
    runtime_checkable,
)

from cached_property import (
    cached_property,
)

from minos.common import (
    classname,
)

from ...context import (
    SagaContext,
)
from ...exceptions import (
    EmptySagaStepException,
    MultipleOnExecuteException,
    MultipleOnFailureException,
    UndefinedOnExecuteException,
)
from ..operations import (
    SagaOperation,
    SagaOperationDecorator,
)
from ..types import (
    LocalCallback,
)
from .abc import (
    SagaStep,
    SagaStepDecoratorMeta,
    SagaStepDecoratorWrapper,
)


@runtime_checkable
class LocalSagaStepDecoratorWrapper(SagaStepDecoratorWrapper, Protocol):
    """Local Saga Step Decorator Wrapper class."""

    meta: LocalSagaStepDecoratorMeta
    on_failure: type[SagaOperationDecorator[LocalCallback]]
    __call__: LocalCallback


class LocalSagaStepDecoratorMeta(SagaStepDecoratorMeta):
    """Local Saga Step Decorator Meta class."""

    _definition: LocalSagaStep
    _on_failure: Optional[LocalCallback]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._on_failure = None

    @cached_property
    def definition(self):
        """Get the step definition.

        :return: A ``SagaStep`` instance.
        """
        self._definition.on_execute(self._inner)
        if self._on_failure is not None:
            self._definition.on_failure(self._on_failure)
        return self._definition

    @cached_property
    def on_failure(self) -> SagaOperationDecorator:
        """Get the on failure decorator.

        :return: A ``SagaOperationDecorator`` type.
        """
        # noinspection PyTypeChecker
        return partial(SagaOperationDecorator[LocalCallback], step_meta=self, attr_name="_on_failure")


class LocalSagaStep(SagaStep):
    """Local Saga Step class."""

    def __init__(
        self,
        on_execute: Optional[Union[LocalCallback, SagaOperation[LocalCallback]]] = None,
        on_failure: Optional[Union[LocalCallback, SagaOperation[LocalCallback]]] = None,
        **kwargs,
    ):

        if on_execute is not None and not isinstance(on_execute, SagaOperation):
            on_execute = SagaOperation(on_execute)
        if on_failure is not None and not isinstance(on_failure, SagaOperation):
            on_failure = SagaOperation(on_failure)

        self.on_execute_operation = on_execute
        self.on_failure_operation = on_failure

        super().__init__(**kwargs)

    @classmethod
    def _from_raw(cls, raw: dict[str, Any]) -> LocalSagaStep:
        raw["on_execute"] = SagaOperation.from_raw(raw["on_execute"])
        raw["on_failure"] = SagaOperation.from_raw(raw["on_failure"])

        return cls(**raw)

    def __call__(self, func: LocalCallback) -> LocalSagaStepDecoratorWrapper:
        """Decorate the given function.

        :param func: The function to be decorated.
        :return: The decorated function.
        """
        meta = LocalSagaStepDecoratorMeta(func, self)
        func.meta = meta
        func.on_failure = meta.on_failure
        return func

    def on_execute(
        self, operation: Union[SagaOperation, LocalCallback], parameters: Optional[SagaContext] = None, **kwargs
    ) -> LocalSagaStep:
        """On execute method.

        :param operation: The callback function to be called.
        :param parameters: A mapping of named parameters to be passed to the callback.
        :param kwargs: A set of named arguments to be passed to the callback. ``parameters`` has priority if it is not
            ``None``.
        :return: A ``self`` reference.
        """
        if self.on_execute_operation is not None:
            raise MultipleOnExecuteException()

        if not isinstance(operation, SagaOperation):
            operation = SagaOperation(operation, parameters, **kwargs)

        self.on_execute_operation = operation

        return self

    def on_failure(
        self, operation: Union[SagaOperation, LocalCallback], parameters: Optional[SagaContext] = None, **kwargs
    ) -> LocalSagaStep:
        """On failure method.

        :param operation: The callback function to be called.
        :param parameters: A mapping of named parameters to be passed to the callback.
        :param kwargs: A set of named arguments to be passed to the callback. ``parameters`` has priority if it is not
            ``None``.
        :return: A ``self`` reference.
        """
        if self.on_failure_operation is not None:
            raise MultipleOnFailureException()

        if not isinstance(operation, SagaOperation):
            operation = SagaOperation(operation, parameters, **kwargs)

        self.on_failure_operation = operation

        return self

    def validate(self) -> None:
        """Check if the step is valid.

        :return: This method does not return anything, but raises an exception if the step is not valid.
        """
        if self.on_execute_operation is None and self.on_failure_operation is None:
            raise EmptySagaStepException()

        if self.on_execute_operation is None:
            raise UndefinedOnExecuteException()

    @property
    def raw(self) -> dict[str, Any]:
        """Generate a raw representation of the instance.

        :return: A ``dict`` instance.
        """
        return {
            "cls": classname(type(self)),
            "order": self.order,
            "on_execute": None if self.on_execute_operation is None else self.on_execute_operation.raw,
            "on_failure": None if self.on_failure_operation is None else self.on_failure_operation.raw,
        }

    def __iter__(self) -> Iterable:
        yield from (
            self.order,
            self.on_execute_operation,
            self.on_failure_operation,
        )



from .operations import (
    SagaOperation,
    SagaOperationDecorator,
)
from .saga import (
    Saga,
    SagaDecoratorMeta,
    SagaDecoratorWrapper,
)
from .steps import (
    ConditionalSagaStep,
    ConditionalSagaStepDecoratorMeta,
    ConditionalSagaStepDecoratorWrapper,
    ElseThenAlternative,
    ElseThenAlternativeDecoratorMeta,
    ElseThenAlternativeDecoratorWrapper,
    IfThenAlternative,
    IfThenAlternativeDecoratorMeta,
    IfThenAlternativeDecoratorWrapper,
    LocalSagaStep,
    LocalSagaStepDecoratorMeta,
    LocalSagaStepDecoratorWrapper,
    RemoteSagaStep,
    RemoteSagaStepDecoratorMeta,
    RemoteSagaStepDecoratorWrapper,
    SagaStep,
    SagaStepDecoratorMeta,
    SagaStepDecoratorWrapper,
)
from .types import (
    ConditionCallback,
    LocalCallback,
    RequestCallBack,
    ResponseCallBack,
)



from collections.abc import (
    Callable,
)
from typing import (
    Awaitable,
    Optional,
    Union,
)

from ..context import (
    SagaContext,
)
from ..messages import (
    SagaRequest,
    SagaResponse,
)

RequestCallBack = Callable[[SagaContext, ...], Union[SagaRequest, Awaitable[SagaRequest]]]
ResponseCallBack = Callable[
    [SagaContext, SagaResponse, ...], Union[Union[Exception, SagaContext], Awaitable[Union[Exception, SagaContext]]]
]
LocalCallback = Callable[[SagaContext, ...], Union[Optional[SagaContext], Awaitable[Optional[SagaContext]]]]
ConditionCallback = Callable[[SagaContext, ...], Union[bool, Awaitable[bool]]]



from typing import (
    Optional,
)

from ...context import (
    SagaContext,
)
from ...definitions import (
    ResponseCallBack,
    SagaOperation,
)
from ...exceptions import (
    ExecutorException,
    SagaFailedExecutionStepException,
)
from ...messages import (
    SagaResponse,
)
from .abc import (
    Executor,
)


class ResponseExecutor(Executor):
    """Response Executor class."""

    # noinspection PyUnusedLocal,PyMethodOverriding
    async def exec(
        self,
        operation: Optional[SagaOperation[ResponseCallBack]],
        context: SagaContext,
        response: SagaResponse,
        *args,
        **kwargs
    ) -> SagaContext:
        """Execute the operation.

        :param operation: Operation to be executed.
        :param context: Actual execution context.
        :param response: SagaResponse containing the response content.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: An updated context instance.
        """
        if operation is None:
            return context

        try:
            context = SagaContext(**context)  # Needed to avoid mutability issues.
            context = await super().exec(operation, context, response)
        except ExecutorException as exc:
            raise SagaFailedExecutionStepException(exc.exception)

        if isinstance(context, Exception):
            raise SagaFailedExecutionStepException(context)

        return context



import inspect
import sys
from typing import (
    Any,
    Callable,
)
from uuid import (
    UUID,
)

from minos.transactions import (
    TransactionEntry,
)

from ...definitions import (
    SagaOperation,
)
from ...exceptions import (
    ExecutorException,
)


class Executor:
    """Executor class."""

    def __init__(self, execution_uuid: UUID, *args, **kwargs):
        self.execution_uuid = execution_uuid

    async def exec(self, operation: SagaOperation, *args, **kwargs) -> Any:
        """Execute the given operation.

        :param operation: The operation to be executed.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: The execution response.
        """

        if operation.parameterized:
            kwargs = {**operation.parameters, **kwargs}

        try:
            func = self._get_method(operation.callback)
        except Exception:  # FIXME
            func = operation.callback

        return await self.exec_function(func, *args, **kwargs)

    def _get_method(self, func) -> type:
        cls = self._get_cls(func)
        obj = cls()
        method = getattr(obj, func.__name__)
        return method

    @staticmethod
    def _get_cls(func):
        vals = vars(sys.modules[func.__module__])
        for attr in func.__qualname__.split(".")[:-1]:
            vals = vals[attr]
        return vals

    async def exec_function(self, func: Callable, *args, **kwargs) -> Any:
        """Execute a function.

        :param func: Function to be executed.
        :param args: Additional positional arguments to the function.
        :param kwargs: Additional named arguments to the function.
        :return: The ``func`` result.
        """
        try:
            async with TransactionEntry(uuid=self.execution_uuid, autocommit=False):
                result = func(*args, **kwargs)
                if inspect.isawaitable(result):
                    result = await result
        except Exception as exc:
            raise ExecutorException(exc)
        return result



from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Union,
)
from uuid import (
    UUID,
)

from minos.common import (
    SetupMixin,
)

from ..saga import (
    SagaExecution,
)


class SagaExecutionRepository(SetupMixin, ABC):
    """Saga Execution Repository class."""

    async def store(self, execution: SagaExecution) -> None:
        """Store an execution.

        :param execution: Execution to be stored.
        :return: This method does not return anything.
        """
        return await self._store(execution)

    async def _store(self, execution: SagaExecution) -> None:
        raise NotImplementedError

    async def load(self, uuid: Union[str, UUID]) -> SagaExecution:
        """Load the saga execution stored on the given key.

        :param uuid: The key to identify the execution.
        :return: A ``SagaExecution`` instance.
        """
        if not isinstance(uuid, UUID):
            uuid = UUID(uuid)
        return await self._load(uuid)

    @abstractmethod
    async def _load(self, uuid: UUID) -> SagaExecution:
        raise NotImplementedError

    async def delete(self, uuid: Union[SagaExecution, str, UUID]) -> None:
        """Delete the reference of the given key.

        :param uuid: Execution key to be deleted.
        :return: This method does not return anything.
        """
        if isinstance(uuid, SagaExecution):
            uuid = uuid.uuid
        if not isinstance(uuid, UUID):
            uuid = UUID(uuid)
        return await self._delete(uuid)

    async def _delete(self, key: UUID) -> None:
        raise NotImplementedError



"""Remote Step Definitions module."""

from __future__ import (
    annotations,
)

from collections.abc import (
    Iterable,
)
from functools import (
    partial,
)
from typing import (
    Any,
    Optional,
    Protocol,
    Union,
    runtime_checkable,
)

from cached_property import (
    cached_property,
)

from minos.common import (
    classname,
)

from ...context import (
    SagaContext,
)
from ...exceptions import (
    EmptySagaStepException,
    MultipleOnErrorException,
    MultipleOnExecuteException,
    MultipleOnFailureException,
    MultipleOnSuccessException,
    UndefinedOnExecuteException,
)
from ..operations import (
    SagaOperation,
    SagaOperationDecorator,
)
from ..types import (
    RequestCallBack,
    ResponseCallBack,
)
from .abc import (
    SagaStep,
    SagaStepDecoratorMeta,
    SagaStepDecoratorWrapper,
)


@runtime_checkable
class RemoteSagaStepDecoratorWrapper(SagaStepDecoratorWrapper, Protocol):
    """Remote Saga Step Decorator Wrapper class."""

    meta: RemoteSagaStepDecoratorMeta
    on_success: type[SagaOperationDecorator[ResponseCallBack]]
    on_error: type[SagaOperationDecorator[ResponseCallBack]]
    on_failure: type[SagaOperationDecorator[RequestCallBack]]
    __call__: RequestCallBack


class RemoteSagaStepDecoratorMeta(SagaStepDecoratorMeta):
    """Remote Saga Step Decorator Meta class."""

    _definition: RemoteSagaStep
    _on_success: Optional[ResponseCallBack]
    _on_error: Optional[ResponseCallBack]
    _on_failure: Optional[RequestCallBack]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._on_success = None
        self._on_error = None
        self._on_failure = None

    @cached_property
    def definition(self):
        """Get the step definition.

        :return: A ``SagaStep`` instance.
        """
        self._definition.on_execute(self._inner)
        if self._on_success is not None:
            self._definition.on_success(self._on_success)
        if self._on_error is not None:
            self._definition.on_error(self._on_error)
        if self._on_failure is not None:
            self._definition.on_failure(self._on_failure)
        return self._definition

    @cached_property
    def on_success(self) -> SagaOperationDecorator:
        """Get the on success decorator.

        :return: A ``SagaOperationDecorator`` type.
        """
        # noinspection PyTypeChecker
        return partial(SagaOperationDecorator[ResponseCallBack], step_meta=self, attr_name="_on_success")

    @cached_property
    def on_error(self) -> SagaOperationDecorator:
        """Get the on error decorator.

        :return: A ``SagaOperationDecorator`` type.
        """
        # noinspection PyTypeChecker
        return partial(SagaOperationDecorator[ResponseCallBack], step_meta=self, attr_name="_on_error")

    @cached_property
    def on_failure(self) -> SagaOperationDecorator:
        """Get the on failure decorator.

        :return: A ``SagaOperationDecorator`` type.
        """
        # noinspection PyTypeChecker
        return partial(SagaOperationDecorator[RequestCallBack], step_meta=self, attr_name="_on_failure")


class RemoteSagaStep(SagaStep):
    """Remote Saga Step class."""

    def __init__(
        self,
        on_execute: Optional[Union[RequestCallBack, SagaOperation[RequestCallBack]]] = None,
        on_success: Optional[Union[ResponseCallBack, SagaOperation[ResponseCallBack]]] = None,
        on_error: Optional[Union[ResponseCallBack, SagaOperation[ResponseCallBack]]] = None,
        on_failure: Optional[Union[RequestCallBack, SagaOperation[RequestCallBack]]] = None,
        **kwargs,
    ):

        if on_execute is not None and not isinstance(on_execute, SagaOperation):
            on_execute = SagaOperation(on_execute)
        if on_failure is not None and not isinstance(on_failure, SagaOperation):
            on_failure = SagaOperation(on_failure)
        if on_success is not None and not isinstance(on_success, SagaOperation):
            on_success = SagaOperation(on_success)
        if on_error is not None and not isinstance(on_error, SagaOperation):
            on_error = SagaOperation(on_error)

        self.on_execute_operation = on_execute
        self.on_failure_operation = on_failure
        self.on_success_operation = on_success
        self.on_error_operation = on_error

        super().__init__(**kwargs)

    @classmethod
    def _from_raw(cls, raw: dict[str, Any]) -> RemoteSagaStep:
        raw["on_execute"] = SagaOperation.from_raw(raw["on_execute"])
        raw["on_failure"] = SagaOperation.from_raw(raw["on_failure"])
        raw["on_success"] = SagaOperation.from_raw(raw["on_success"])
        raw["on_error"] = SagaOperation.from_raw(raw["on_error"])

        return cls(**raw)

    def __call__(self, func: RequestCallBack) -> RemoteSagaStepDecoratorWrapper:
        """Decorate the given function.

        :param func: The function to be decorated.
        :return: The decorated function.
        """
        meta = RemoteSagaStepDecoratorMeta(func, self)
        func.meta = meta
        func.on_success = meta.on_success
        func.on_error = meta.on_error
        func.on_failure = meta.on_failure
        return func

    def on_execute(
        self,
        operation: Union[SagaOperation[RequestCallBack], RequestCallBack],
        parameters: Optional[SagaContext] = None,
        **kwargs,
    ) -> RemoteSagaStep:
        """On execute method.

        :param operation: The callback function to be called.
        :param parameters: A mapping of named parameters to be passed to the callback.
        :param kwargs: A set of named arguments to be passed to the callback. ``parameters`` has priority if it is not
            ``None``.
        :return: A ``self`` reference.
        """
        if self.on_execute_operation is not None:
            raise MultipleOnExecuteException()

        if not isinstance(operation, SagaOperation):
            operation = SagaOperation(operation, parameters, **kwargs)

        self.on_execute_operation = operation

        return self

    def on_failure(
        self,
        operation: Union[SagaOperation[RequestCallBack], RequestCallBack],
        parameters: Optional[SagaContext] = None,
        **kwargs,
    ) -> RemoteSagaStep:
        """On failure method.

        :param operation: The callback function to be called.
        :param parameters: A mapping of named parameters to be passed to the callback.
        :param kwargs: A set of named arguments to be passed to the callback. ``parameters`` has priority if it is not
            ``None``.
        :return: A ``self`` reference.
        """
        if self.on_failure_operation is not None:
            raise MultipleOnFailureException()

        if not isinstance(operation, SagaOperation):
            operation = SagaOperation(operation, parameters, **kwargs)

        self.on_failure_operation = operation

        return self

    def on_success(
        self,
        operation: Union[SagaOperation[ResponseCallBack], ResponseCallBack],
        parameters: Optional[SagaContext] = None,
        **kwargs,
    ) -> RemoteSagaStep:
        """On success method.

        :param operation: The callback function to be called.
        :param parameters: A mapping of named parameters to be passed to the callback.
        :param kwargs: A set of named arguments to be passed to the callback. ``parameters`` has priority if it is not
            ``None``.
        :return: A ``self`` reference.
        """
        if self.on_success_operation is not None:
            raise MultipleOnSuccessException()

        if not isinstance(operation, SagaOperation):
            operation = SagaOperation(operation, parameters, **kwargs)

        self.on_success_operation = operation

        return self

    def on_error(
        self,
        operation: Union[SagaOperation[ResponseCallBack], ResponseCallBack],
        parameters: Optional[SagaContext] = None,
        **kwargs,
    ) -> RemoteSagaStep:
        """On error method.

        :param operation: The callback function to be called.
        :param parameters: A mapping of named parameters to be passed to the callback.
        :param kwargs: A set of named arguments to be passed to the callback. ``parameters`` has priority if it is not
            ``None``.
        :return: A ``self`` reference.
        """
        if self.on_error_operation is not None:
            raise MultipleOnErrorException()

        if not isinstance(operation, SagaOperation):
            operation = SagaOperation(operation, parameters, **kwargs)

        self.on_error_operation = operation

        return self

    def validate(self) -> None:
        """Check if the step is valid.

        :return: This method does not return anything, but raises an exception if the step is not valid.
        """
        if (
            self.on_execute_operation is None
            and self.on_failure_operation is None
            and self.on_success_operation is None
            and self.on_error_operation is None
        ):
            raise EmptySagaStepException()

        if self.on_execute_operation is None:
            raise UndefinedOnExecuteException()

    @property
    def raw(self) -> dict[str, Any]:
        """Generate a raw representation of the instance.

        :return: A ``dict`` instance.
        """
        return {
            "cls": classname(type(self)),
            "order": self.order,
            "on_execute": None if self.on_execute_operation is None else self.on_execute_operation.raw,
            "on_failure": None if self.on_failure_operation is None else self.on_failure_operation.raw,
            "on_success": None if self.on_success_operation is None else self.on_success_operation.raw,
            "on_error": None if self.on_error_operation is None else self.on_error_operation.raw,
        }

    def __iter__(self) -> Iterable:
        yield from (
            self.order,
            self.on_execute_operation,
            self.on_failure_operation,
            self.on_success_operation,
            self.on_error_operation,
        )



"""Operation module."""

from __future__ import (
    annotations,
)

from collections.abc import (
    Callable,
    Iterable,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Optional,
    TypeVar,
    Union,
)

if TYPE_CHECKING:
    from .steps import SagaStepDecoratorMeta

from minos.common import (
    classname,
    import_module,
)

from ..context import (
    SagaContext,
)

T = TypeVar("T", bound=Callable)


class SagaOperationDecorator(Generic[T]):
    """Saga Operation Decorator class."""

    def __init__(self, attr_name: str = None, step_meta: SagaStepDecoratorMeta = None, *args, **kwargs):
        if attr_name is None:
            raise ValueError(f"The 'attr_name' must not be {None!r}.")
        if step_meta is None:
            raise ValueError(f"The 'step_meta' must not be {None!r}.")

        self._step_meta = step_meta
        self._attr_name = attr_name

        self._args = args
        self._kwargs = kwargs

    def __call__(self, func: T) -> T:
        operation = SagaOperation(func, *self._args, **self._kwargs)
        setattr(self._step_meta, self._attr_name, operation)
        return func


class SagaOperation(Generic[T]):
    """Saga Step Operation class."""

    def __init__(self, callback: T, parameters: Optional[Union[dict, SagaContext]] = None, **kwargs):
        if parameters is None and len(kwargs):
            parameters = kwargs
        if parameters is not None and not isinstance(parameters, SagaContext):
            parameters = SagaContext(**parameters)

        self.callback = callback
        self.parameters = parameters

    @property
    def raw(self) -> dict[str, Any]:
        """Generate a rew representation of the instance.

        :return: A ``dict`` instance.
        """
        # noinspection PyTypeChecker
        raw = {"callback": classname(self.callback)}

        if self.parameterized:
            raw["parameters"] = self.parameters.avro_str

        return raw

    @property
    def parameterized(self) -> bool:
        """parameterized getter.

        :return: ``True`` if parameters are provided or ``False`` otherwise.
        """
        return self.parameters is not None

    @classmethod
    def from_raw(cls, raw: Optional[Union[dict[str, Any], SagaOperation[T]]], **kwargs) -> Optional[SagaOperation[T]]:
        """Build a new instance from a raw representation.

        :param raw: A raw representation.
        :param kwargs: Additional named arguments.
        :return: A ``SagaStepOperation`` instance if the ``raw`` argument is not ``None``, ``None`` otherwise.
        """
        if raw is None:
            return None

        if isinstance(raw, cls):
            return raw

        current = raw | kwargs
        if isinstance(current["callback"], str):
            current["callback"] = import_module(current["callback"])

        if "parameters" in current:
            current["parameters"] = SagaContext.from_avro_str(current["parameters"])
        return cls(**current)

    def __hash__(self):
        return hash(tuple(self))

    def __eq__(self, other: SagaOperation) -> bool:
        return type(self) == type(other) and tuple(self) == tuple(other)

    def __repr__(self) -> str:
        return f"{type(self).__name__}{tuple(self)}"

    def __iter__(self) -> Iterable:
        yield from (
            self.callback,
            self.parameters,
        )



from .commit import (
    TransactionCommitter,
)
from .executors import (
    Executor,
    LocalExecutor,
    RequestExecutor,
    ResponseExecutor,
)
from .repositories import (
    DatabaseSagaExecutionRepository,
    SagaExecutionDatabaseOperationFactory,
    SagaExecutionRepository,
)
from .runners import (
    SagaRunner,
)
from .saga import (
    SagaExecution,
)
from .status import (
    SagaStatus,
    SagaStepStatus,
)
from .steps import (
    ConditionalSagaStepExecution,
    LocalSagaStepExecution,
    RemoteSagaStepExecution,
    SagaStepExecution,
)



from __future__ import (
    annotations,
)

from enum import (
    Enum,
)
from typing import (
    Union,
)


class SagaStatus(str, Enum):
    """Saga Status class."""

    Created = "created"
    Running = "running"
    Paused = "paused"
    Finished = "finished"
    Errored = "errored"

    @classmethod
    def from_raw(cls, raw: Union[str, SagaStatus]) -> SagaStatus:
        """Build a new instance from raw.

        :param raw: The raw representation of the instance.
        :return: A ``SagaStatus`` instance.
        """
        if isinstance(raw, cls):
            return raw

        # noinspection PyTypeChecker,PyUnresolvedReferences
        return next(obj for obj in cls if obj.value == raw)

    @property
    def raw(self) -> str:
        """Compute the raw representation of the instance.

        :return: A ``str`` instance.
        """
        return self.value


class SagaStepStatus(str, Enum):
    """Saga Step Status class."""

    Created = "created"
    RunningOnExecute = "running-on-execute"
    FinishedOnExecute = "finished-on-execute"
    ErroredOnExecute = "errored-on-execute"
    PausedByOnExecute = "paused-by-on-execute"
    ErroredByOnExecute = "errored-by-on-execute"
    RunningOnSuccess = "running-on-success"
    FinishedOnSuccess = "finished-on-success"
    ErroredOnSuccess = "errored-on-success"
    RunningOnError = "running-on-error"
    FinishedOnError = "finished-on-error"
    ErroredOnError = "errored-on-error"
    Finished = "finished"

    @classmethod
    def from_raw(cls, raw: Union[str, SagaStepStatus]) -> SagaStepStatus:
        """Build a new instance from raw.

        :param raw: The raw representation of the instance.
        :return: A ``SagaStepStatus`` instance.
        """
        if isinstance(raw, cls):
            return raw

        # noinspection PyTypeChecker,PyUnresolvedReferences
        return next(obj for obj in cls if obj.value == raw)

    @property
    def raw(self) -> str:
        """Compute the raw representation of the instance.

        :return: A ``str`` instance.
        """
        return self.value
